


# import typing as tp
import unittest
from hypothesis import given
import numpy as np
from scipy.stats import rankdata

from static_frame.test.property import strategies as sfst
from static_frame.test.test_case import TestCase
from static_frame.core.rank import rank_1d
from static_frame.core.rank import rank_2d



class TestUnit(TestCase):


    @given(sfst.get_array_1d(dtype_group=sfst.DTGroup.NUMERIC, max_size=100))
    def test_rank_1d_ordinal(self, value: np.ndarray) -> None:
        a1 = rankdata(value, method='ordinal')
        a2 = rank_1d(value, method='ordinal', start=1)
        self.assertEqual(a1.tolist(), a2.tolist())

        if len(value):
            a3 = rank_1d(value, method='ordinal')
            self.assertEqual(a3.min(), 0)

            a4 = rank_1d(value, method='ordinal', ascending=False)
            self.assertEqual(a4.min(), 0)

    @given(sfst.get_array_1d(dtype_group=sfst.DTGroup.NUMERIC, max_size=100))
    def test_rank_1d_dense(self, value: np.ndarray) -> None:
        # cannot compare values with NaN as scipy uses quicksort
        if np.isnan(value).any():
            return
        a1 = rankdata(value, method='dense')
        a2 = rank_1d(value, method='dense', start=1)
        self.assertEqual(a1.tolist(), a2.tolist())

        if len(value):
            a3 = rank_1d(value, method='dense')
            self.assertEqual(a3.min(), 0)

            a4 = rank_1d(value, method='dense', ascending=False)
            self.assertEqual(a4.min(), 0)

    @given(sfst.get_array_1d(dtype_group=sfst.DTGroup.NUMERIC, max_size=100))
    def test_rank_1d_min(self, value: np.ndarray) -> None:
        # cannot compare values with NaN as scipy uses quicksort
        if np.isnan(value).any():
            return
        a1 = rankdata(value, method='min')
        a2 = rank_1d(value, method='min', start=1)
        self.assertEqual(a1.tolist(), a2.tolist())

        if len(value):
            a3 = rank_1d(value, method='min')
            self.assertEqual(a3.min(), 0)

            a4 = rank_1d(value, method='min', ascending=False)
            self.assertEqual(a4.min(), 0)


    @given(sfst.get_array_1d(dtype_group=sfst.DTGroup.NUMERIC, max_size=100))
    def test_rank_1d_max(self, value: np.ndarray) -> None:
        # cannot compare values with NaN as scipy uses quicksort
        if np.isnan(value).any():
            return
        a1 = rankdata(value, method='max')
        a2 = rank_1d(value, method='max', start=1)
        self.assertEqual(a1.tolist(), a2.tolist())



    @given(sfst.get_array_1d(dtype_group=sfst.DTGroup.NUMERIC, max_size=100))
    def test_rank_1d_average(self, value: np.ndarray) -> None:
        # cannot compare values with NaN as scipy uses quicksort
        if np.isnan(value).any():
            return
        a1 = rankdata(value, method='average')
        a2 = rank_1d(value, method='mean', start=1)
        self.assertEqual(a1.tolist(), a2.tolist())




    @given(sfst.get_array_2d(dtype_group=sfst.DTGroup.NUMERIC, max_rows=20, max_columns=20))
    def test_rank_2d_ordinal(self, value: np.ndarray) -> None:
        for axis in (0, 1):
            a1 = rankdata(value, method='ordinal', axis=axis)
            a2 = rank_2d(value, method='ordinal', start=1, axis=axis)
            self.assertEqual(a1.tolist(), a2.tolist())

    @given(sfst.get_array_2d(dtype_group=sfst.DTGroup.NUMERIC, max_rows=20, max_columns=20))
    def test_rank_2d_dense(self, value: np.ndarray) -> None:
        # cannot compare values with NaN as scipy uses quicksort
        if np.isnan(value).any():
            return
        for axis in (0, 1):
            a1 = rankdata(value, method='dense', axis=axis)
            a2 = rank_2d(value, method='dense', start=1, axis=axis)
            self.assertEqual(a1.tolist(), a2.tolist())

    @given(sfst.get_array_2d(dtype_group=sfst.DTGroup.NUMERIC, max_rows=20, max_columns=20))
    def test_rank_2d_min(self, value: np.ndarray) -> None:
        # cannot compare values with NaN as scipy uses quicksort
        if np.isnan(value).any():
            return
        for axis in (0, 1):
            a1 = rankdata(value, method='min', axis=axis)
            a2 = rank_2d(value, method='min', start=1, axis=axis)
            self.assertEqual(a1.tolist(), a2.tolist())

    @given(sfst.get_array_2d(dtype_group=sfst.DTGroup.NUMERIC, max_rows=20, max_columns=20))
    def test_rank_2d_max(self, value: np.ndarray) -> None:
        # cannot compare values with NaN as scipy uses quicksort
        if np.isnan(value).any():
            return
        for axis in (0, 1):
            a1 = rankdata(value, method='max', axis=axis)
            a2 = rank_2d(value, method='max', start=1, axis=axis)
            self.assertEqual(a1.tolist(), a2.tolist())

    @given(sfst.get_array_2d(dtype_group=sfst.DTGroup.NUMERIC, max_rows=20, max_columns=20))
    def test_rank_2d_average(self, value: np.ndarray) -> None:
        # cannot compare values with NaN as scipy uses quicksort
        if np.isnan(value).any():
            return
        for axis in (0, 1):
            a1 = rankdata(value, method='average', axis=axis)
            a2 = rank_2d(value, method='mean', start=1, axis=axis)
            self.assertEqual(a1.tolist(), a2.tolist())





if __name__ == '__main__':
    unittest.main()




import unittest
import typing as tp

import numpy as np
import static_frame as sf

from static_frame.test.test_case import TestCase

CHARACTERS_REFERENCE = dict((
        ('count', len),
        ('count<0', lambda x: len(x.loc[x < 0])),
        ('count>0', lambda x: len(x.loc[x > 0])),
        ('count_unique', lambda x: len(x.dropna().unique())),
        ('min', lambda x: x.min()),
        ('max', lambda x: x.max()),
        ('sum', lambda x: x.sum()),
        ('mean', lambda x: x.mean()),
        ('median', lambda x: x.median()),
        ('nanfill', lambda x: x.isna().sum() / len(x)),
        ))

class FieldStatConfig:

    CHARACTERS = tuple((k, v) for k, v in CHARACTERS_REFERENCE.items())

    CHARACTERS_AGGREGATE = tuple((k, v)
            for k, v in CHARACTERS_REFERENCE.items()
            if not (k.startswith('count') or k.startswith('nan')))

    GROUP_BY = 'group_by'
    CATEGORY_A_FIELDS: tp.Iterable[tp.Tuple[str, tp.Any]] = (('a1', None), ('a2', None))
    CATEGORY_B_FIELDS: tp.Iterable[tp.Tuple[str, tp.Any]] = (('b1', None), ('b2', None))

    # store pairs of lable, config attr
    CATEGORY_LABELS = (
            ('category_a', 'CATEGORY_A_FIELDS'),
            ('category_b', 'CATEGORY_B_FIELDS'),
            )



def process(
        frame: sf.Frame,
        config: tp.Type[FieldStatConfig] = FieldStatConfig
        ) -> sf.Frame:
    '''
    Perform statistical analysis, returning a new Frame.
    '''
    # print(frame.columns.display(sf.DisplayConfigs.UNBOUnND))

    def observations(
            fields: tp.Iterable[tp.Tuple[str, tp.Any]],
            ) -> tp.Iterator[sf.Series]:
        for k, group in frame.iter_group_items(config.GROUP_BY):

            def gen() -> tp.Iterator[tp.Tuple[tp.Tuple[str, str], sf.Series]]:
                for field, converter in fields:
                    # the key here becomes columns
                    converter = converter if converter else lambda x: x
                    for label, func in config.CHARACTERS:
                        yield (field, label), func(converter(group[field])) # type: ignore

            # name will become row (index) identifier
            yield sf.Series.from_items(gen(),
                    name=k,
                    index_constructor=sf.IndexHierarchy.from_labels)

    def field_categories() -> tp.Iterator[sf.Frame]:

        for label, attr in config.CATEGORY_LABELS:
            # for each category, get all fields, and process each character, for each field, into one Frame
            fields = getattr(config, attr)


            post = sf.Frame.from_concat(observations(fields))

            # import ipdb; ipdb.set_trace()

            # create more rows with axis config.CHARACTERS over the groups
            def gen() -> tp.Iterator[sf.Frame]:
                for label, func in config.CHARACTERS_AGGREGATE:
                    yield func(post).rename(label) # type: ignore

            post = sf.Frame.from_concat((post, sf.Frame.from_concat(gen())))

            # round float, rotate table, and new index for this level
            # NOTe: after iter_lement apply, we are loosing hierarchucal index
            yield post.iter_element().apply(lambda e: round(e, 3)).T.relabel_level_add(label)

    # vertically stack all Frame for each category
    return sf.Frame.from_concat(field_categories()) # type: ignore



class TestUnit(TestCase):


    def test_process(self) -> None:

        a1 = np.array([.1, -.1, np.nan, .2])
        a2 = np.array([.4, np.nan, .7, -.5])

        records = (
            a1,
            a1 * .2,
            a1 * -.5,
            a1 * 0,
            np.flip(a1),
            np.flip(a1) * .2,
            a2,
            a2 * .2,
            a2 * -.5,
            a2 * 0,
            np.flip(a2),
            np.flip(a2) * .2,
        )

        f = sf.FrameGO.from_records(records, columns=('a1', 'a2', 'b1', 'b2'))
        f['group_by'] = ['x'] * (len(f) // 2) + ['y'] * (len(f) // 2)


        post = process(f)

        self.assertEqual(post.shape, (40, 7))
        self.assertEqual(post.index.depth, 3)

        self.assertEqual(post.iloc[0].to_pairs(),
            (('x', 6.0), ('y', 6.0), ('min', 6.0), ('max', 6.0), ('sum', 12.0), ('mean', 6.0), ('median', 6.0))
            )

        self.assertEqual(post['x'].to_pairs(),
            ((('category_a', 'a1', 'count'), 6.0), (('category_a', 'a1', 'count<0'), 1.0), (('category_a', 'a1', 'count>0'), 4.0), (('category_a', 'a1', 'count_unique'), 6.0), (('category_a', 'a1', 'min'), -0.05), (('category_a', 'a1', 'max'), 0.2), (('category_a', 'a1', 'sum'), 0.31), (('category_a', 'a1', 'mean'), 0.052), (('category_a', 'a1', 'median'), 0.03), (('category_a', 'a1', 'nanfill'), 0.0), (('category_a', 'a2', 'count'), 6.0), (('category_a', 'a2', 'count<0'), 2.0), (('category_a', 'a2', 'count>0'), 1.0), (('category_a', 'a2', 'count_unique'), 4.0), (('category_a', 'a2', 'min'), -0.1), (('category_a', 'a2', 'max'), 0.05), (('category_a', 'a2', 'sum'), -0.07), (('category_a', 'a2', 'mean'), -0.018), (('category_a', 'a2', 'median'), -0.01), (('category_a', 'a2', 'nanfill'), 0.333), (('category_b', 'b1', 'count'), 6.0), (('category_b', 'b1', 'count<0'), 2.0), (('category_b', 'b1', 'count>0'), 0.0), (('category_b', 'b1', 'count_unique'), 2.0), (('category_b', 'b1', 'min'), -0.1), (('category_b', 'b1', 'max'), -0.02), (('category_b', 'b1', 'sum'), -0.12), (('category_b', 'b1', 'mean'), -0.06), (('category_b', 'b1', 'median'), -0.06), (('category_b', 'b1', 'nanfill'), 0.667), (('category_b', 'b2', 'count'), 6.0), (('category_b', 'b2', 'count<0'), 1.0), (('category_b', 'b2', 'count>0'), 4.0), (('category_b', 'b2', 'count_unique'), 6.0), (('category_b', 'b2', 'min'), -0.1), (('category_b', 'b2', 'max'), 0.2), (('category_b', 'b2', 'sum'), 0.26), (('category_b', 'b2', 'mean'), 0.043), (('category_b', 'b2', 'median'), 0.03), (('category_b', 'b2', 'nanfill'), 0.0))
        )

        self.assertAlmostEqualValues(post.sum().values.tolist(),
            [58.466999999999985, 61.121, 48.06699999999999, 71.52100000000004, 119.587, 59.794000000000004, 59.794000000000004])

if __name__ == '__main__':
    unittest.main()




import unittest

from static_frame.core.style_config import StyleConfig
from static_frame.core.style_config import style_config_css_factory
from static_frame.core.frame import Frame
from static_frame.test.test_case import TestCase


class TestUnit(TestCase):

    #---------------------------------------------------------------------------
    def test_style_config_a(self) -> None:

        sc = StyleConfig()
        self.assertEqual(sc.frame(), '')
        self.assertEqual(sc.apex(3, (1,1)), ('3', ''))
        self.assertEqual(sc.values(3, (1,1)), ('3', ''))
        self.assertEqual(sc.index(3), ('3', ''))
        self.assertEqual(sc.columns(3), ('3', ''))


    def test_style_config_css_factory_a(self) -> None:
        f = Frame()
        sc = StyleConfig(f)
        self.assertTrue(style_config_css_factory(sc, f) is sc)


if __name__ == '__main__':
    unittest.main()






import os
import typing as tp

import numpy as np

import static_frame as sf
from static_frame.test.test_case import Timer



# def func_b(frame: Frame) -> Frame:
#     for row in frame.iter_series():
#         pass
#         # if row[10] > 1000: # type: ignore
#     return frame


# def main() -> None:

#     f1 = Frame(np.arange(100000000).reshape(1000000, 100), name='a')
#     f2 = Frame(np.arange(100000000).reshape(1000000, 100), name='b')
#     f3 = Frame(np.arange(100000000).reshape(1000000, 100), name='c')
#     f4 = Frame(np.arange(100000000).reshape(1000000, 100), name='d')
#     f5 = Frame(np.arange(100000000).reshape(1000000, 100), name='e')
#     f6 = Frame(np.arange(100000000).reshape(1000000, 100), name='f')
#     f7 = Frame(np.arange(100000000).reshape(1000000, 100), name='g')
#     f8 = Frame(np.arange(100000000).reshape(1000000, 100), name='h')



#     @timer #type: ignore
#     def a1() -> None:
#         batch1 = Batch.from_frames((f1, f2, f3, f4, f5, f6, f7, f8))
#         batch2 = (batch1 * 100).sum()
#         _ = tuple(batch2.items())

#     @timer #type: ignore
#     def a2() -> None:
#         batch1 = Batch.from_frames((f1, f2, f3, f4, f5, f6, f7, f8), max_workers=6, use_threads=True)
#         batch2 = (batch1 * 100).sum()
#         post = dict(batch2.items())

#     a1()
#     a2()

#     @timer #type: ignore
#     def b1() -> None:
#         batch1 = Batch.from_frames((f1, f2, f3, f4, f5, f6, f7, f8))
#         batch2 = batch1.apply(func_b)
#         _ = tuple(batch2.items())

#     @timer #type: ignore
#     def b2() -> None:
#         batch1 = Batch.from_frames((f1, f2, f3, f4, f5, f6, f7, f8), max_workers=8, use_threads=False, chunksize=2)
#         batch2 = batch1.apply(func_b)
#         _ = tuple(batch2.items())

#     # b1()
    # b2()


def bus_batch_streaming() -> None:

    TypeIterFrameItems = tp.Iterator[tp.Tuple[tp.Hashable, sf.Frame]]

    # do a bunch of processing on large Frames while maintaining single frame overhead

    def origin_data() -> TypeIterFrameItems:
        for label, i in ((chr(i), i) for i in range(65, 75)): # A, B, ...
            f = sf.Frame(np.arange(100000).reshape(1000, 100) * i, name=label)
            yield label, f

    # a StoreConfig is bundle of read/write config in a single object
    config = sf.StoreConfig(include_index=True,
            include_columns=True,
            index_depth=1,
            columns_depth=1)

    # incrementally generate and write data one Frame at at ime
    sf.Batch(origin_data()).to_zip_parquet('/tmp/pq-stg-01.zip', config=config)

    # we can read the same data store into Bus
    # settting max_persist to 1 keeps only 1 frame in memory
    src_bus = sf.Bus.from_zip_parquet('/tmp/pq-stg-01.zip', max_persist=1, config=config)

    def proc_data(bus: sf.Bus) -> TypeIterFrameItems:
        for label in bus.keys():
            f = bus[label]
            f_post = f * .00001
            yield label, f_post

    # incrementally read, process data, and write data
    # note that we do not necessarily need to write this intermediate step; we might chain a number of processors together
    sf.Batch(proc_data(src_bus)).to_zip_parquet('/tmp/pq-stg-02.zip', config=config)

    # a function that reads through the derived data and produces single in-memory result
    def derive_characteristic(bus: sf.Bus) -> sf.Series:
        def gen() -> tp.Iterator[tp.Tuple[tp.Hashable, float]]:
            for label in bus.keys():
                f = bus[label]
                yield label, f.mean().mean()

        return sf.Series.from_items(gen())

    post_bus = sf.Bus.from_zip_parquet('/tmp/pq-stg-02.zip', max_persist=1, config=config)
    print(derive_characteristic(post_bus))

    # <Series>
    # <Index>
    # A        32.499674999999996
    # B        32.99967
    # C        33.49966500000001
    # D        33.999660000000006
    # E        34.499655000000004
    # F        34.99965
    # G        35.49964500000001
    # H        35.99964000000001
    # I        36.499635000000005
    # J        36.999629999999996
    # <<U1>    <float64>

    # if we have family of functions that process items pairs that do not need random access, we can chain them together without writing to an intermediary

    def proc_data_alt(items: TypeIterFrameItems) -> TypeIterFrameItems:
        for label, f in items:
            f_post = f * .00001
            yield label, f_post

    def derive_characteristic_alt(items: TypeIterFrameItems) -> sf.Series:
        def gen() -> tp.Iterator[tp.Tuple[tp.Hashable, float]]:
            for label, f in items:
                yield label, f.mean().mean()
        return sf.Series.from_items(gen())

    # now we can use a Batch to get sequential processing, and start with the origin data set
    batch = sf.Batch.from_zip_parquet('/tmp/pq-stg-01.zip', config=config)
    print(derive_characteristic_alt(proc_data_alt(batch.items())))

    # <Series>
    # <Index>
    # A        32.499674999999996
    # B        32.99967
    # C        33.49966500000001
    # D        33.999660000000006
    # E        34.499655000000004
    # F        34.99965
    # G        35.49964500000001
    # H        35.99964000000001
    # I        36.499635000000005
    # J        36.999629999999996
    # <<U1>    <float64>


def bus_batch_demo() -> None:
    # field definitions
    # https://www.ndbc.noaa.gov/measdes.shtml

    # full data set
    # https://www.ndbc.noaa.gov/view_text_file.php?filename=46222h2018.txt.gz&dir=data/historical/stdmet/


    # multi-table storage: XLSX, HDF5, SQLite, zipped bundles
    # how can we easily read, write, and process these

    # Batch and the Bus: two series-like containers of Frames
    # Frames can have different shapes and types: not a Panel
    # Bus: random access to Frame, writing to multi-table formats, lazy loading
    # Batch: lazy, sequential processor of Frame, with support for reading/writing to multi-table formats

    sp2018 = sf.Frame.from_tsv('/tmp/san_pedro.txt')

    sp2018 = sf.Frame.from_tsv('/tmp/san_pedro.txt', dtypes={'datetime': 'datetime64[m]'})

    sp2018.shape
    # (16824, 5)

    sp2018 = sp2018.set_index('datetime', index_constructor=sf.IndexMinute, drop=True)

    # sp2018.loc['2018-01'].shape
    # (1419, 4)

    yms = sf.IndexYearMonth.from_year_month_range('2018-01', '2018-12')

    sp_per_month = [sp2018.loc[ym].rename(str(ym)) for ym in yms]

    [f.shape for f in sp_per_month]
    # [(1419, 4), (1295, 4), (1460, 4), (1411, 4), (1431, 4), (1418, 4), (1471, 4), (1160, 4), (1426, 4), (1471, 4), (1420, 4), (1442, 4)]

    sp_per_month[0].to_xlsx('/tmp/sp_2018_01.xlsx')


    # [TALK] introduce Bus: how to create a multi-sheet workbook
    # pandas: pd.ExcelWriter and pd.HDFStore

    sf.Bus.interface # from frames


    b1 = sf.Bus.from_frames(sp_per_month)
    b1.to_xlsx('/tmp/sp_2018.xlsx')


    # [TALK] to do this in a single line:
    # but note that this will full-create the Bus before writing

    sf.Bus.from_frames(sp2018.loc[ym].rename(str(ym)) for ym in yms).to_xlsx('/tmp/sp_2018.xlsx')


    # [TALK] what other formats can we export to
    # show interface, constructors and exporters
    b1.interface

    b1.to_sqlite('/tmp/sp_2018.sqlite')


    # [TALK] value of parquet and pickle

    b1.to_zip_pickle('/tmp/sp_2018.zip')


    # [TALK] Bus is Series like: can select using loc, iloc, etc
    # limit in that the index has to be strings: need to use as labels in output formats

    b1['2018-08']
    b1['2018-08':] #type: ignore
    b1.iloc[[0, 5, 10]]
    b1[b1.index.isin(('2018-01', '2018-09'))]


    # [TALK] What about reading: we can lazily read from these output formats
    sf.Bus.interface


    b2 = sf.Bus.from_zip_pickle('/tmp/sp_2018.zip')

    # notice that we have FrameDeferred
    # when access Frames, the simply get loaded as the are selected:

    b2['2018-08']
    b2 # show only one Frame loaded

    b2.iloc[[0, 5, 10]]
    b2 # show we have loaded additional frames


    # what if you have 600 Frames and you do not want the stored frames to all load
    b3 = sf.Bus.from_zip_pickle('/tmp/sp_2018.zip', max_persist=2)

    for label in b3.index:
        print(b3[label].shape)
        print(b3)


    #---------------------------------------------------------------------------
    # there are other situations where we need to deal with multiple Frames at as as single unit

    import pandas as pd
    df = pd.read_csv('/tmp/san_pedro.txt', sep='\t')
    df['year_mo'] = df['datetime'].apply(lambda d: d[:7])

    # what happens when we do a group-by in Pandas

    df.groupby('year_mo')
    # <pandas.core.groupby.generic.DataFrameGroupBy object at 0x7fb3263e7090>

    # can iterate
    [k for k, v in df.groupby('year_mo')]
# ['2018-01', '2018-02', '2018-03', '2018-04', '2018-05', '2018-06', '2018-07', '2018-08', '2018-09', '2018-10', '2018-11', '2018-12']

    # can take the max of each group in one line
    df.groupby('year_mo')['WVHT'].max()

    # how does this work: some operations are deferred, some result in containers
    df.groupby('year_mo')['WVHT']
    # <pandas.core.groupby.generic.SeriesGroupBy object at 0x7fb32e1b7fd0>

    # but what if we want to do the same thing with Frames we created without a groupby, or by some other means?


    #---------------------------------------------------------------------------
    # the Batch

    batch1 = sf.Batch.from_frames(sp_per_month)

    # can call to_frame to realize all data with a hierarchical index:
    sf.Batch.from_frames(sp_per_month).to_frame()


    # can do the same thing as pandas, but have to explicitly call to_frame() to get a result
    batch1['WVHT'].max().to_frame()


    # Batch are like generators: one-time use, one-time iteration. If we look at batch1 again, we see it is empty
    batch1

    # the reason we have to explicitly call to_frame as that there is no limit on how many operations you can do on the Batch; all are deferred until final iteration or `to_frame` call
    # here, we convert meters to feet before printing getting result

    (sf.Batch.from_frames(sp_per_month)['WVHT'].max() * 3.28084).to_frame()


    # what if we want to do this on a group by?
    # SF's iter_group is simply an iterator of groups, we can feed it into a Batch


    # lets create a fresh FrameGO and add a year mo column
    sp2018 = sf.FrameGO.from_tsv('/tmp/san_pedro.txt', dtypes={'datetime': 'datetime64[m]'})
    sp2018 = sp2018.set_index('datetime', index_constructor=sf.IndexMinute, drop=False)
    sp2018['year_mo'] = sp2018['datetime'].astype('datetime64[M]') #type: ignore

    sp2018.iter_group_items('year_mo')

    # we can feed the iterator of pairs of label, frame to Batch, then process
    sf.Batch(sp2018.iter_group_items('year_mo'))['WVHT'].max().to_frame()


    # any time we have iterators of pairs of label, Frame, we can use a Batch
    # for example, what if we want to iterate on windows

    tuple(sp2018.iter_window_items(size=100))[0]

    # We can feed those same pairs into a Batch and get the rolling mean

    sf.Batch(sp2018.iter_window_items(size=100))['WVHT'].mean().to_frame()

    # Again, we can convert meters to feet:

    (sf.Batch(sp2018.iter_window_items(size=100))['WVHT'].mean() * 3.28084).to_frame()

    # These operations are pretty fast, but we can potential optimize them by using multi threading or multi processing

    (sf.Batch(sp2018.iter_window_items(size=100), max_workers=6, use_threads=True)['WVHT'].mean() * 3.28084).to_frame()


    # there is also a generic apply method to perform arbitrary functions
    sf.Batch(sp2018.iter_window_items(size=100, step=100)).apply(lambda f: f.loc[f['DPD'].loc_max(), ['WVHT', 'DPD']]).to_frame()


    # what if we want to write read or write from a multi-table format
    # because the Batch is a lazy sequential processor, this is actually a pipeline that is processing one table at time
    # memory overhead is one table at a time


    sf.Batch.from_xlsx('/tmp/sp_2018.xlsx').apply(lambda f: f.assign['WVHT'](f['WVHT'] * 3.28084)).to_xlsx('/tmp/sp_2018_ft.xlsx')




def bus_aggregate() -> None:
    import frame_fixtures as ff
    from itertools import zip_longest

    # create the "total index"
    index = sf.IndexDate.from_date_range('2019-01-12', '2019-12-31')

    # create dummy data index by the total index and with columns
    src = ff.parse(f's({len(index)},4)').relabel(index=index, columns=tuple('abcd'))

    chunk_size = 7
    window_size = 12

    # create a bus by chunking src
    def items_frames() -> tp.Iterator[sf.Frame]:
        starts = range(0, len(index), chunk_size)
        ends = range(starts[1], len(index), chunk_size)

        for start, end in zip_longest(starts, ends, fillvalue=len(index)):
            f = src.iloc[start:end]
            yield f.rename(str(f.index.iloc[0]))

    bus = sf.Bus.from_frames(items_frames())

    # create a series of total index to bus label
    def items_map() -> tp.Iterator[tp.Tuple[np.datetime64, str]]:
        for f in bus.values:
            for dt in f.index:
                yield dt, f.name

    ref = sf.Series.from_items(items_map())

    # using the ref, we can select and concat from any start, end
    def get_slice(start: np.datetime64, end: np.datetime64) -> sf.Frame:
        bus_labels = ref[start: end].unique()
        return sf.Frame.from_concat(bus[bus_labels].values).loc[start: end] #type: ignore


    # selecting an arbitrary window
    fsub = get_slice(index.iloc[3], index.iloc[100])

    # creating a batch for windowed processing
    def items_window() -> tp.Iterator[tp.Tuple[np.datetime64, sf.Frame]]:
        for window in index.to_series().iter_window(size=window_size):
            fsub = get_slice(window.values[0], window.values[-1]) #type: ignore
            yield fsub.index[-1], fsub

    post = sf.Batch(items_window()).mean().to_frame()

    # comparing src to batch-extracted windows
    post_alt = sf.Batch(src.iter_window_items(size=window_size)).mean().to_frame()
    assert post.equals(post_alt)





#-------------------------------------------------------------------------------

# https://www.kaggle.com/unsdsn/world-happiness
#

# Other possibility:
# https://www.kaggle.com/borismarjanovic/price-volume-data-for-all-us-stocks-etfs

#-------------------------------------------------------------------------------

def main() -> None:
    fp = '/home/ariza/Downloads/archive.zip'

    config = sf.StoreConfig(index_depth=0, label_decoder=int, label_encoder=str)

    bus = sf.Bus.from_zip_csv(fp, config=config)

    def normalize(f: sf.Frame) -> sf.Frame:
        fields = ['country', 'score', 'rank', 'gdp']

        def relabel(label: str) -> str:
            for field in fields:
                if field in label.lower():
                    return field
            return label

        return f.relabel(columns=relabel)[fields].set_index(fields[0], drop=True) #type: ignore

    bus = sf.Batch.from_zip_csv(fp, config=config).apply(normalize).to_bus()

    # selection
    batch = sf.Batch(bus.items())
    quilt = sf.Quilt(bus, axis=0, retain_labels=True)


def tables() -> None:
    name = 'Higher-Order Containers'
    columns = ('Bus', 'Batch', 'Yarn', 'Quilt')
    records_items = (
    ('Presented ndim',        (1,         2,        1,                  2)),
    ('Approximate Interface', ('Series',  'Frame', 'Series',           'Frame')),
    ('Composes',              ('n Frame', '',      'm Bus of n Frame', '1 Bus of n Frame of shape (x, y)')),
    ('Presented shape',       ('(n,)',    '',      '(mn,)',            '(xn, y) or (x, yn)' )),
    )

    f = sf.Frame.from_records_items(records_items, columns=columns, name=name)
    # print(f)
    config = sf.DisplayConfig(cell_max_width=200, type_show=False)
    print(f.name)
    print(f.to_markdown(config=config))

    # name = 'Constructors & Exporters'
    # columns = ('Constructor', 'Exporter') #type: ignore
    # records = (
    #     ('from_zip_tsv', 'to_zip_tsv',),
    #     ('from_zip_csv', 'to_zip_csv',),
    #     ('from_zip_pickle', 'to_zip_pickle',),
    #     ('from_zip_parquet', 'to_zip_parquet',),
    #     ('from_xlsx',  'to_xlsx'),
    #     ('from_sqlite',  'to_sqlite'),
    #     ('from_hdf5',  'to_hdf5'),
    #     )

    # f = sf.Frame.from_records(records, columns=columns, name=name)
    # # print(f)
    # print(f.name)
    # print(f.to_markdown(sf.DisplayConfig(include_index=False, type_show=False)))



def stocks_write() -> None:

    t = Timer()
    d = '/home/ariza/Downloads/archive/Stocks'
    fps = ((fn, os.path.join(d, fn)) for fn in os.listdir(d))
    items = ((fn.replace('.us.txt', ''), sf.Frame.from_csv(fp, index_depth=1)) for fn, fp in fps if os.path.getsize(fp))
    sf.Batch(items).to_zip_pickle('/tmp/stocks.zip')
    print(t)

def stocks() -> None:
    t = Timer()
    bus = sf.Bus.from_zip_pickle('/tmp/stocks.zip')[:]
    print(t, 'done loading')

    t = Timer()
    post = sf.Batch(bus.items())[['Open', 'Close']].loc_max().to_frame()
    print(t, 'serial')


    #>> quilt.loc[sf.HLoc[:, '2017-11-10']]


if __name__ == '__main__':
    # stocks()
    tables()
    # main()







    # def format(f: sf.Frame) -> sf.Frame:
    #     field_to_name = {
    #         'country': ('Country', 'Country or region'),
    #         'score': ('Score', 'Happiness.Score', 'Happiness Score'),
    #         'rank': ('Overall rank', 'Happiness.Rank', 'Happiness Rank'),
    #         'gdppc': ('Economy (GDP per Capita)', 'Economy..GDP.per.Capita.', 'GDP per capita'),
    #     }

    #     def get_fields(label):
    #         for field, names in field_to_name.items():
    #             if label in names:
    #                 return field
    #         return label

    #     return f.relabel(columns=get_fields)[list(field_to_name)].set_index(
    #             'country', drop=True)



import unittest
from static_frame.test.test_case import TestCase
from static_frame.core.frame import Frame
from static_frame.test.test_case import temp_file
from static_frame.core.index_auto import IndexAutoFactory

class TestUnit(TestCase):

    def test_exceed_columns(self) -> None:

        f1 = Frame.from_element('x', index='x', columns=range(16384))

        with temp_file('.xlsx') as fp:

            with self.assertRaises(RuntimeError):
                # with the index, the limit is exceeded
                f1.to_xlsx(fp, include_index=True)


            f1.to_xlsx(fp, include_index=False)
            f2 = Frame.from_xlsx(fp, index_depth=0, columns_depth=1)
            # need to remove index on original for appropriate comparison
            self.assertEqualFrames(f1.relabel(index=IndexAutoFactory), f2)


    def test_exceed_rows(self) -> None:

        f1 = Frame.from_element('x', index=range(1048576), columns='x')

        with temp_file('.xlsx') as fp:

            with self.assertRaises(RuntimeError):
                # with the index, the limit is exceeded
                f1.to_xlsx(fp, include_columns=True)

            # NOTE: it takes almost 60s to write this file, so we will skip testing it
            # f1.to_xlsx(fp, include_columns=False)
            # f2 = Frame.from_xlsx(fp, index_depth=1, columns_depth=0)
            # # need to remove index on original for appropriate comparison
            # self.assertEqualFrames(f1.relabel(columns=IndexAutoFactory), f2)



if __name__ == '__main__':
    unittest.main()




import timeit
import typing as tp
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import ProcessPoolExecutor


import numpy as np

import static_frame as sf




class PerfTest:
    NUMBER = 50

    def __init__(self, size: int, groups: int):
        self.fixture = np.arange(0, size) % groups
        self.groups = range(groups)

    def run(self):
        raise NotImplementedError()


class Serial(PerfTest):
    def run(self):
        selections = []
        for i in self.groups:
            v = self.fixture == i
            selections.append(v)

class SerialNoCopy(PerfTest):
    def run(self):
        v = np.empty(len(self.fixture), dtype=bool)
        for i in self.groups:
            np.equal(self.fixture, i, out=v)

def func(args):
    return np.equal(*args)

class ThreadPool(PerfTest):
    def run(self):
        futures = []
        with ThreadPoolExecutor(max_workers=100) as executor:
            for i in self.groups:
                futures.append(executor.submit(func, (self.fixture, i)))
            results = [f.result() for f in futures]

# class ProcessPool(PerfTest):
#     def run(self):
#         futures = []
#         with ProcessPoolExecutor() as executor:
#             for i in self.groups:
#                 futures.append(executor.submit(func, (self.fixture, i)))
#             results = [f.result() for f in futures]


#-------------------------------------------------------------------------------
def get_format():

    name_root_last = None
    name_root_count = 0

    def format(key: tp.Tuple[tp.Any, str], v: object) -> str:
        nonlocal name_root_last
        nonlocal name_root_count

        if isinstance(v, float):
            if np.isnan(v):
                return ''
            return str(round(v, 4))
        if isinstance(v, (bool, np.bool_)):
            if v:
                return HexColor.format_terminal('green', str(v))
            return HexColor.format_terminal('orange', str(v))

        return str(v)

    return format


def run_test():
    records = []

    for size, groups in (
        (1_000, 2),
        (1_000, 100),
        (100_000, 2),
        (100_000, 100),
        (100_000, 1_000),
        ):
        for cls in (
                Serial,
                SerialNoCopy,
                ThreadPool,
                # ProcessPool,
                ):
            runner = cls(size, groups)
            record = [cls.__name__, cls.NUMBER, size, groups]
            result = timeit.timeit(
                    f'runner.run()',
                    globals=locals(),
                    number=cls.NUMBER)
            record.append(result)
            records.append(record)

    f = sf.FrameGO.from_records(records,
            columns=('name', 'number', 'size', 'groups', 'time')
            )

    display = f.iter_element_items().apply(get_format())

    config = sf.DisplayConfig(
            cell_max_width_leftmost=np.inf,
            cell_max_width=np.inf,
            type_show=False,
            display_rows=200,
            include_index=False,
            )
    print(display.display(config))
    # import ipdb; ipdb.set_trace()

if __name__ == '__main__':
    run_test()







import unittest
import datetime

import frame_fixtures as ff
import numpy as np

from static_frame.test.test_case import TestCase
from static_frame.core.bus import Bus
from static_frame.core.index_auto import IndexAutoFactory
from static_frame.core.display_config import DisplayConfig
from static_frame.core.yarn import Yarn
from static_frame.core.frame import Frame
from static_frame.test.test_case import temp_file
from static_frame.core.exception import ErrorInitYarn
from static_frame.core.exception import ErrorInitIndex
from static_frame.core.exception import ErrorInitSeries
from static_frame.core.index_datetime import IndexDate
from static_frame import ILoc
from static_frame import HLoc
from static_frame.core.exception import RelabelInvalid
from static_frame.core.index_hierarchy import IndexHierarchy
from static_frame.core.series import Series



class TestUnit(TestCase):

    #---------------------------------------------------------------------------
    def test_yarn_init_a(self) -> None:

        with self.assertRaises(ErrorInitSeries):
            Yarn(np.array([3, 4]))

    def test_yarn_init_b(self) -> None:

        f1 = ff.parse('s(4,4)|v(int,float)').rename('f1')
        f2 = ff.parse('s(4,4)|v(str)').rename('f2')
        f3 = ff.parse('s(4,4)|v(bool)').rename('f3')
        b1 = Bus.from_frames((f1, f2, f3))

        f4 = ff.parse('s(4,4)|v(int,float)').rename('f4')
        f5 = ff.parse('s(4,4)|v(str)').rename('f5')
        b2 = Bus.from_frames((f4, f5))

        y1 = Yarn((b1, b2), index=tuple('abcde'))
        self.assertEqual(y1.index.values.tolist(), list('abcde'))
        self.assertEqual(y1[['a', 'c', 'e']].shape, (3,))

        y2 = Yarn((b1, b2))
        self.assertEqual(y2.index.values.tolist(), list(range(5)))
        self.assertEqual(y2[2:].shape, (3,))

        y3 = Yarn((b2,), index=('2021-01-01', '2021-02-15'), index_constructor=IndexDate)
        self.assertEqual(y3.index.__class__, IndexDate)
        self.assertEqual(y3.index.values.tolist(), [datetime.date(2021, 1, 1), datetime.date(2021, 2, 15)])

        with self.assertRaises(ErrorInitYarn):
            y4 = Yarn((b2,), index=range(5))

    def test_yarn_init_c(self) -> None:

        with self.assertRaises(ErrorInitYarn):
            Yarn((ff.parse('s(2,2)'),))

        with self.assertRaises(ErrorInitYarn):
            Yarn(Series((ff.parse('s(2,2)'),), dtype=object))


    def test_yarn_init_d(self) -> None:

        with self.assertRaises(ErrorInitYarn):
            Yarn(Series(np.array((False, True))))


    #---------------------------------------------------------------------------
    def test_yarn_from_buses_a(self) -> None:

        f1 = ff.parse('s(4,4)|v(int,float)').rename('f1')
        f2 = ff.parse('s(4,4)|v(str)').rename('f2')
        f3 = ff.parse('s(4,4)|v(bool)').rename('f3')
        b1 = Bus.from_frames((f1, f2, f3), name='a')

        f4 = ff.parse('s(4,4)|v(int,float)').rename('f4')
        f5 = ff.parse('s(4,4)|v(str)').rename('f5')
        b2 = Bus.from_frames((f4, f5), name='b')


        y1 = Yarn.from_buses((b1, b2), retain_labels=True)
        self.assertEqual(len(y1), 5)
        self.assertEqual(y1.index.shape, (5, 2))
        self.assertEqual(y1.shape, (5,))
        self.assertEqual(y1.size, 5)
        self.assertEqual(y1.dtype, object)
        self.assertEqual(y1.ndim, 1)

        y3 = y1[('a', 'f2'):] #type: ignore
        self.assertEqual(y3.shape, (4,))

        y2 = Yarn.from_buses((b1, b2), retain_labels=False)
        self.assertEqual(len(y2), 5)
        self.assertEqual(y2.index.shape, (5,))
        self.assertEqual(y1.shape, (5,))
        self.assertEqual(y1.size, 5)
        self.assertEqual(y1.dtype, object)
        self.assertEqual(y1.ndim, 1)


    def test_yarn_from_buses_b(self) -> None:

        f1 = ff.parse('s(4,4)|v(int,float)').rename('f1')
        f2 = ff.parse('s(4,4)|v(str)').rename('f2')
        f3 = ff.parse('s(4,4)|v(bool)').rename('f3')
        b1 = Bus.from_frames((f1, f2, f3), name='a')

        f4 = ff.parse('s(4,4)|v(int,float)').rename('f4')
        f5 = ff.parse('s(4,4)|v(str)').rename('f5')
        b2 = Bus.from_frames((f4, f5), name='b')

        f6 = ff.parse('s(2,4)|v(int,float)').rename('f6')
        f7 = ff.parse('s(4,2)|v(str)').rename('f7')
        b3 = Bus.from_frames((f6, f7), name='c')

        y1 = Yarn.from_buses((b1, b2, b3), retain_labels=True)
        self.assertEqual(y1.shape, (7,))
        self.assertEqual(len(y1), 7)

    #---------------------------------------------------------------------------
    def test_yarn_max_persist(self) -> None:
        f1 = ff.parse('s(4,2)').rename('f1')
        f2 = ff.parse('s(4,5)').rename('f2')
        f3 = ff.parse('s(2,2)').rename('f3')
        f4 = ff.parse('s(2,8)').rename('f4')
        f5 = ff.parse('s(4,4)').rename('f5')
        f6 = ff.parse('s(6,4)').rename('f6')

        b1 = Bus.from_frames((f1, f2, f3))
        b2 = Bus.from_frames((f4, f5, f6))

        with temp_file('.zip') as fp1, temp_file('.zip') as fp2:
            b1.to_zip_pickle(fp1)
            b2.to_zip_pickle(fp2)


            bus_a = Bus.from_zip_pickle(fp1, max_persist=1).rename('a')
            bus_b = Bus.from_zip_pickle(fp2, max_persist=1).rename('b')

            y1 = Yarn.from_buses((bus_a, bus_b), retain_labels=False)
            self.assertEqual(y1.nbytes, 0)
            self.assertEqual(y1.status['loaded'].sum(), 0)

            self.assertEqual(y1['f2'].shape, (4, 5))
            self.assertEqual(y1['f6'].shape, (6, 4))
            self.assertEqual(y1.nbytes, 352)
            self.assertEqual(y1.status['loaded'].sum(), 2)

            self.assertEqual(y1.shapes.to_pairs(),
                    (('f1', None), ('f2', (4, 5)), ('f3', None), ('f4', None), ('f5', None), ('f6', (6, 4)))
                    )
            self.assertEqual(y1.mloc.isna().sum(), 4)
            self.assertEqual((y1.dtypes == float).sum().sum(), 9)

    #---------------------------------------------------------------------------
    def test_yarn_from_concat_a(self) -> None:
        f1 = ff.parse('s(4,4)|v(int,float)').rename('f1')
        f2 = ff.parse('s(4,4)|v(str)').rename('f2')
        f3 = ff.parse('s(4,4)|v(bool)').rename('f3')
        b1 = Bus.from_frames((f1, f2, f3), name='a')

        f4 = ff.parse('s(4,4)|v(int,float)').rename('f4')
        f5 = ff.parse('s(4,4)|v(str)').rename('f5')
        b2 = Bus.from_frames((f4, f5), name='b')

        f6 = ff.parse('s(2,4)|v(int,float)').rename('f6')
        f7 = ff.parse('s(4,2)|v(str)').rename('f7')
        b3 = Bus.from_frames((f6, f7), name='c')

        y1 = Yarn.from_concat((Yarn.from_buses((b1,), retain_labels=True), Yarn.from_buses((b2, b3), retain_labels=True)))
        self.assertEqual(y1.shape, (7,))
        self.assertEqual(y1.index.values.tolist(),
                [['a', 'f1'], ['a', 'f2'], ['a', 'f3'], ['b', 'f4'], ['b', 'f5'], ['c', 'f6'], ['c', 'f7']]
                )


    def test_yarn_from_concat_b(self) -> None:
        f1 = ff.parse('s(4,2)').rename('f1')
        f2 = ff.parse('s(4,5)').rename('f2')
        f3 = ff.parse('s(2,2)').rename('f3')
        f4 = ff.parse('s(2,8)').rename('f4')
        f5 = ff.parse('s(4,4)').rename('f5')
        f6 = ff.parse('s(6,4)').rename('f6')

        b1 = Bus.from_frames((f1, f2, f3))
        b2 = Bus.from_frames((f4, f5, f6))

        with temp_file('.zip') as fp1, temp_file('.zip') as fp2:
            b1.to_zip_pickle(fp1)
            b2.to_zip_pickle(fp2)

            bus_a = Bus.from_zip_pickle(fp1, max_persist=1).rename('a')
            bus_b = Bus.from_zip_pickle(fp2, max_persist=1).rename('b')

            y1 = Yarn.from_concat((Yarn.from_buses((bus_a,), retain_labels=True), Yarn.from_buses((bus_b,), retain_labels=True)))

            y2 = Yarn.from_concat((y1, y1), index=IndexAutoFactory)

            self.assertEqual(y2[3].shape, (2, 8))
            self.assertEqual(y2[0].shape, (4, 2))
            self.assertEqual(y2[5].shape, (6, 4))

            y3 = y2.iloc[4:]
            self.assertEqual(y3.shape, (8,))


    def test_yarn_from_concat_c(self) -> None:
        f1 = ff.parse('s(4,2)').rename('f1')
        f2 = ff.parse('s(4,5)').rename('f2')

        with self.assertRaises(NotImplementedError):
            Yarn.from_concat((f1, f2))

    #---------------------------------------------------------------------------
    def test_yarn_reversed_a(self) -> None:
        f1 = ff.parse('s(4,4)|v(int,float)').rename('f1')
        f2 = ff.parse('s(4,4)|v(str)').rename('f2')
        f3 = ff.parse('s(4,4)|v(bool)').rename('f3')
        b1 = Bus.from_frames((f1, f2, f3), name='a')

        f4 = ff.parse('s(4,4)|v(int,float)').rename('f4')
        f5 = ff.parse('s(4,4)|v(str)').rename('f5')
        b2 = Bus.from_frames((f4, f5), name='b')

        f6 = ff.parse('s(2,4)|v(int,float)').rename('f6')
        f7 = ff.parse('s(4,2)|v(str)').rename('f7')
        b3 = Bus.from_frames((f6, f7), name='c')

        y1 = Yarn.from_buses((b1, b2, b3), retain_labels=False)
        self.assertEqual(tuple(reversed(y1)), ('f7', 'f6', 'f5', 'f4', 'f3', 'f2', 'f1'))

    #---------------------------------------------------------------------------
    def test_yarn_size_a(self) -> None:
        f1 = ff.parse('s(4,4)|v(int,float)').rename('f1')
        f2 = ff.parse('s(4,4)|v(str)').rename('f2')
        f3 = ff.parse('s(4,4)|v(bool)').rename('f3')
        b1 = Bus.from_frames((f1, f2, f3), name='a')

        f4 = ff.parse('s(4,4)|v(int,float)').rename('f4')
        f5 = ff.parse('s(4,4)|v(str)').rename('f5')
        b2 = Bus.from_frames((f4, f5), name='b')

        y1 = Yarn.from_buses((b1, b2), retain_labels=False, name='foo')
        self.assertEqual(y1.size, 5)

    #---------------------------------------------------------------------------
    def test_yarn_index_a(self) -> None:
        f1 = ff.parse('s(4,4)|v(int,float)').rename('f1')
        f2 = ff.parse('s(4,4)|v(str)').rename('f2')
        f3 = ff.parse('s(4,4)|v(bool)').rename('f3')
        b1 = Bus.from_frames((f1, f2, f3), name='a')

        f4 = ff.parse('s(4,4)|v(int,float)').rename('f4')
        f5 = ff.parse('s(4,4)|v(str)').rename('f5')
        b2 = Bus.from_frames((f4, f5), name='b')

        y1 = Yarn.from_buses((b1, b2), retain_labels=False, name='foo')
        self.assertEqual(y1.index.values.tolist(), ['f1', 'f2', 'f3', 'f4', 'f5'])


    #---------------------------------------------------------------------------
    def test_yarn_rename_a(self) -> None:
        f1 = ff.parse('s(4,4)|v(int,float)').rename('f1')
        f2 = ff.parse('s(4,4)|v(str)').rename('f2')
        f3 = ff.parse('s(4,4)|v(bool)').rename('f3')
        b1 = Bus.from_frames((f1, f2, f3), name='a')

        f4 = ff.parse('s(4,4)|v(int,float)').rename('f4')
        f5 = ff.parse('s(4,4)|v(str)').rename('f5')
        b2 = Bus.from_frames((f4, f5), name='b')

        f6 = ff.parse('s(2,4)|v(int,float)').rename('f6')
        f7 = ff.parse('s(4,2)|v(str)').rename('f7')
        b3 = Bus.from_frames((f6, f7), name='c')

        y1 = Yarn.from_buses((b1, b2, b3), retain_labels=False, name='foo')
        self.assertEqual(y1.name, 'foo')
        y2 = y1.rename('bar')
        self.assertEqual(y2.name, 'bar')

    #---------------------------------------------------------------------------
    def test_yarn_loc_a(self) -> None:
        f1 = ff.parse('s(4,4)|v(int,float)').rename('f1')
        f2 = ff.parse('s(4,4)|v(str)').rename('f2')
        f3 = ff.parse('s(4,4)|v(bool)').rename('f3')
        b1 = Bus.from_frames((f1, f2, f3), name='a')

        f4 = ff.parse('s(4,4)|v(int,float)').rename('f4')
        f5 = ff.parse('s(4,4)|v(str)').rename('f5')
        b2 = Bus.from_frames((f4, f5), name='b')

        f6 = ff.parse('s(2,4)|v(int,float)').rename('f6')
        f7 = ff.parse('s(4,2)|v(str)').rename('f7')
        b3 = Bus.from_frames((f6, f7), name='c')

        y1 = Yarn.from_buses((b1, b2, b3), retain_labels=False, name='foo')
        self.assertEqual(y1.loc['f4'].shape, (4, 4))
        self.assertEqual(y1.loc['f4':].shape, (4,)) #type: ignore
        self.assertEqual(y1.loc[['f2', 'f7']].shape, (2,))
        self.assertEqual(y1.loc[y1.index.via_str.startswith('f3')].shape, (1,))

    def test_yarn_loc_b(self) -> None:
        f1 = ff.parse('s(4,4)|v(int,float)').rename('f1')
        f2 = ff.parse('s(4,4)|v(str)').rename('f2')
        f3 = ff.parse('s(4,4)|v(bool)').rename('f3')
        b1 = Bus.from_frames((f1, f2, f3), name='a')

        f4 = ff.parse('s(4,4)|v(int,float)').rename('f4')
        f5 = ff.parse('s(4,4)|v(str)').rename('f5')
        b2 = Bus.from_frames((f4, f5), name='b')

        f6 = ff.parse('s(2,4)|v(int,float)').rename('f6')
        f7 = ff.parse('s(4,2)|v(str)').rename('f7')
        b3 = Bus.from_frames((f6, f7), name='c')

        y1 = Yarn.from_buses((b1, b2, b3), retain_labels=False, name='foo')
        y2 = y1.loc[y1.index.via_re('[26]').search()]
        self.assertEqual(y2.index.values.tolist(), ['f2', 'f6'])
        self.assertEqual(y2.shapes.to_pairs(),
                (('f2', (4, 4)), ('f6', (2, 4))))
        self.assertEqual(y2.name, 'foo')

    def test_yarn_loc_c(self) -> None:
        f1 = ff.parse('s(4,4)|v(int,float)').rename('f1')
        f2 = ff.parse('s(4,4)|v(str)').rename('f2')
        f3 = ff.parse('s(4,4)|v(bool)').rename('f3')
        b1 = Bus.from_frames((f1, f2, f3), name='a')

        f4 = ff.parse('s(4,4)|v(int,float)').rename('f4')
        f5 = ff.parse('s(4,4)|v(str)').rename('f5')
        b2 = Bus.from_frames((f4, f5), name='b')

        f6 = ff.parse('s(2,4)|v(int,float)').rename('f6')
        f7 = ff.parse('s(4,2)|v(str)').rename('f7')
        b3 = Bus.from_frames((f6, f7), name='c')

        y1 = Yarn.from_buses((b1, b2, b3), retain_labels=True, name='foo')
        y2 = y1.loc[[('a', 'f3'), ('b', 'f5'), ('c', 'f6')]]
        self.assertEqual(y2.shapes.to_pairs(),
                ((('a', 'f3'), (4, 4)), (('b', 'f5'), (4, 4)), (('c', 'f6'), (2, 4))))


    def test_yarn_loc_d(self) -> None:
        f1 = ff.parse('s(4,4)|v(int,float)').rename('f1')
        f2 = ff.parse('s(4,4)|v(str)').rename('f2')
        f3 = ff.parse('s(4,4)|v(bool)').rename('f3')
        b1 = Bus.from_frames((f1, f2, f3), name='a')

        f4 = ff.parse('s(4,4)|v(int,float)').rename('f4')
        f5 = ff.parse('s(4,4)|v(str)').rename('f5')
        b2 = Bus.from_frames((f4, f5), name='b')

        f6 = ff.parse('s(2,4)|v(int,float)').rename('f6')
        f7 = ff.parse('s(4,2)|v(str)').rename('f7')
        b3 = Bus.from_frames((f6, f7), name='c')

        y1 = Yarn.from_buses((b1, b2, b3), retain_labels=False, name='foo')
        y2 = y1.loc[['f7', 'f3']]
        self.assertEqual(y2.shapes.to_pairs(),
                (('f7', (4, 2)), ('f3', (4, 4)))
                )
        with self.assertRaises(ErrorInitIndex):
            y2 = y1.loc[['f1', 'f7', 'f3']]

    def test_yarn_loc_e(self) -> None:
        f1 = ff.parse('s(4,4)|v(int,float)').rename('f1')
        f2 = ff.parse('s(4,4)|v(str)').rename('f2')
        f3 = ff.parse('s(4,4)|v(bool)').rename('f3')
        b1 = Bus.from_frames((f1, f2, f3), name='a')

        f4 = ff.parse('s(4,4)|v(int,float)').rename('f4')
        f5 = ff.parse('s(4,4)|v(str)').rename('f5')
        b2 = Bus.from_frames((f4, f5), name='b')

        f6 = ff.parse('s(2,4)|v(int,float)').rename('f6')
        f7 = ff.parse('s(4,2)|v(str)').rename('f7')
        b3 = Bus.from_frames((f6, f7), name='c')

        y1 = Yarn.from_buses((b1, b2, b3), retain_labels=False, name='foo')
        y2 = y1['f2':'f6'] #type: ignore
        self.assertEqual(y2.shapes.to_pairs(),
                (('f2', (4, 4)), ('f3', (4, 4)), ('f4', (4, 4)), ('f5', (4, 4)), ('f6', (2, 4))))
        self.assertEqual(y2['f5'].to_pairs(),
                ((0, ((0, 'zjZQ'), (1, 'zO5l'), (2, 'zEdH'), (3, 'zB7E'))), (1, ((0, 'zaji'), (1, 'zJnC'), (2, 'zDdR'), (3, 'zuVU'))), (2, ((0, 'ztsv'), (1, 'zUvW'), (2, 'zkuW'), (3, 'zmVj'))), (3, ((0, 'z2Oo'), (1, 'z5l6'), (2, 'zCE3'), (3, 'zr4u')))))


    def test_yarn_loc_f(self) -> None:
        f1 = ff.parse('s(4,4)|v(int,float)').rename('f1')
        f2 = ff.parse('s(3,4)|v(str)').rename('f2')
        f3 = ff.parse('s(4,5)|v(bool)').rename('f3')
        b1 = Bus.from_frames((f1, f2, f3), name='a')

        f4 = ff.parse('s(4,8)|v(int,float)').rename('f4')
        f5 = ff.parse('s(2,3)|v(str)').rename('f5')
        b2 = Bus.from_frames((f4, f5), name='b')

        f6 = ff.parse('s(2,4)|v(int,float)').rename('f6')
        f7 = ff.parse('s(4,2)|v(str)').rename('f7')
        b3 = Bus.from_frames((f6, f7), name='c')

        y1 = Yarn.from_buses((b1, b2, b3), retain_labels=True, name='foo')
        y2 = y1.loc[[ILoc[2], ('c', 'f6')]]
        self.assertEqual(y2.shapes.to_pairs(),
                ((('a', 'f3'), (4, 5)), (('c', 'f6'), (2, 4))))

        y3 = y1.loc[HLoc[['a', 'c']]]
        self.assertEqual(y3.shapes.to_pairs(),
                ((('a', 'f1'), (4, 4)), (('a', 'f2'), (3, 4)), (('a', 'f3'), (4, 5)), (('c', 'f6'), (2, 4)), (('c', 'f7'), (4, 2))))

    #---------------------------------------------------------------------------
    def test_yarn_iloc_a(self) -> None:
        f1 = ff.parse('s(4,4)|v(int,float)').rename('f1')
        f2 = ff.parse('s(4,4)|v(str)').rename('f2')
        f3 = ff.parse('s(4,4)|v(bool)').rename('f3')
        b1 = Bus.from_frames((f1, f2, f3), name='a')

        f4 = ff.parse('s(4,4)|v(int,float)').rename('f4')
        f5 = ff.parse('s(4,4)|v(str)').rename('f5')
        b2 = Bus.from_frames((f4, f5), name='b')

        f6 = ff.parse('s(2,4)|v(int,float)').rename('f6')
        f7 = ff.parse('s(4,2)|v(str)').rename('f7')
        b3 = Bus.from_frames((f6, f7), name='c')

        y1 = Yarn.from_buses((b1, b2, b3), retain_labels=False, name='foo')
        self.assertEqual(y1.iloc[3].shape, (4, 4))
        self.assertEqual(y1.iloc[3:].shape, (4,))
        self.assertEqual(y1.iloc[[1, 6]].shape, (2,))
        self.assertEqual(y1.iloc[y1.index.via_str.startswith('f3')].shape, (1,))


    #---------------------------------------------------------------------------
    def test_yarn_keys_a(self) -> None:
        f1 = ff.parse('s(4,4)|v(int,float)').rename('f1')
        f2 = ff.parse('s(4,4)|v(str)').rename('f2')
        f3 = ff.parse('s(4,4)|v(bool)').rename('f3')
        b1 = Bus.from_frames((f1, f2, f3), name='a')

        f4 = ff.parse('s(4,4)|v(int,float)').rename('f4')
        f5 = ff.parse('s(4,4)|v(str)').rename('f5')
        b2 = Bus.from_frames((f4, f5), name='b')

        f6 = ff.parse('s(2,4)|v(int,float)').rename('f6')
        f7 = ff.parse('s(4,2)|v(str)').rename('f7')
        b3 = Bus.from_frames((f6, f7), name='c')

        y1 = Yarn.from_buses((b1, b2, b3), retain_labels=False, name='foo')
        self.assertEqual(tuple(y1.keys()), ('f1', 'f2', 'f3', 'f4', 'f5', 'f6', 'f7'))


    #---------------------------------------------------------------------------
    def test_yarn_iter_a(self) -> None:
        f1 = ff.parse('s(4,4)|v(int,float)').rename('f1')
        f2 = ff.parse('s(4,4)|v(str)').rename('f2')
        f3 = ff.parse('s(4,4)|v(bool)').rename('f3')
        b1 = Bus.from_frames((f1, f2, f3), name='a')

        f4 = ff.parse('s(4,4)|v(int,float)').rename('f4')
        f5 = ff.parse('s(4,4)|v(str)').rename('f5')
        b2 = Bus.from_frames((f4, f5), name='b')

        f6 = ff.parse('s(2,4)|v(int,float)').rename('f6')
        f7 = ff.parse('s(4,2)|v(str)').rename('f7')
        b3 = Bus.from_frames((f6, f7), name='c')

        y1 = Yarn.from_buses((b1, b2, b3), retain_labels=False, name='foo')
        self.assertEqual(next(iter(y1)), 'f1')


    #---------------------------------------------------------------------------
    def test_yarn_contains_a(self) -> None:
        f1 = ff.parse('s(4,4)|v(int,float)').rename('f1')
        f2 = ff.parse('s(4,4)|v(str)').rename('f2')
        f3 = ff.parse('s(4,4)|v(bool)').rename('f3')
        b1 = Bus.from_frames((f1, f2, f3), name='a')

        f4 = ff.parse('s(4,4)|v(int,float)').rename('f4')
        f5 = ff.parse('s(4,4)|v(str)').rename('f5')
        b2 = Bus.from_frames((f4, f5), name='b')

        f6 = ff.parse('s(2,4)|v(int,float)').rename('f6')
        f7 = ff.parse('s(4,2)|v(str)').rename('f7')
        b3 = Bus.from_frames((f6, f7), name='c')

        y1 = Yarn.from_buses((b1, b2, b3), retain_labels=False, name='foo')
        self.assertTrue('f6' in y1)


    #---------------------------------------------------------------------------
    def test_yarn_get_a(self) -> None:
        f1 = ff.parse('s(4,4)|v(int,float)').rename('f1')
        f2 = ff.parse('s(4,4)|v(str)').rename('f2')
        f3 = ff.parse('s(4,4)|v(bool)').rename('f3')
        b1 = Bus.from_frames((f1, f2, f3), name='a')

        f4 = ff.parse('s(4,4)|v(int,float)').rename('f4')
        f5 = ff.parse('s(4,4)|v(str)').rename('f5')
        b2 = Bus.from_frames((f4, f5), name='b')

        f6 = ff.parse('s(2,4)|v(int,float)').rename('f6')
        f7 = ff.parse('s(4,2)|v(str)').rename('f7')
        b3 = Bus.from_frames((f6, f7), name='c')

        y1 = Yarn.from_buses((b1, b2, b3), retain_labels=False, name='foo')
        self.assertTrue(y1.get('f2').equals(f2))
        self.assertEqual(y1.get('f99'), None)


    #---------------------------------------------------------------------------
    def test_yarn_head_a(self) -> None:
        f1 = ff.parse('s(4,4)|v(int,float)').rename('f1')
        f2 = ff.parse('s(4,4)|v(str)').rename('f2')
        f3 = ff.parse('s(4,4)|v(bool)').rename('f3')
        b1 = Bus.from_frames((f1, f2, f3), name='a')

        f4 = ff.parse('s(4,4)|v(int,float)').rename('f4')
        f5 = ff.parse('s(4,4)|v(str)').rename('f5')
        b2 = Bus.from_frames((f4, f5), name='b')

        f6 = ff.parse('s(2,4)|v(int,float)').rename('f6')
        f7 = ff.parse('s(4,2)|v(str)').rename('f7')
        b3 = Bus.from_frames((f6, f7), name='c')

        y1 = Yarn.from_buses((b1, b2, b3), retain_labels=False, name='foo')
        self.assertEqual(y1.head(2).shape, (2,))
        self.assertEqual(tuple(y1.head(2).keys()), ('f1', 'f2'))


    #---------------------------------------------------------------------------
    def test_yarn_tail_a(self) -> None:
        f1 = ff.parse('s(4,4)|v(int,float)').rename('f1')
        f2 = ff.parse('s(4,4)|v(str)').rename('f2')
        f3 = ff.parse('s(4,4)|v(bool)').rename('f3')
        b1 = Bus.from_frames((f1, f2, f3), name='a')

        f4 = ff.parse('s(4,4)|v(int,float)').rename('f4')
        f5 = ff.parse('s(4,4)|v(str)').rename('f5')
        b2 = Bus.from_frames((f4, f5), name='b')

        f6 = ff.parse('s(2,4)|v(int,float)').rename('f6')
        f7 = ff.parse('s(4,2)|v(str)').rename('f7')
        b3 = Bus.from_frames((f6, f7), name='c')

        y1 = Yarn.from_buses((b1, b2, b3), retain_labels=False, name='foo')
        self.assertEqual(y1.tail(2).shape, (2,))
        self.assertEqual(tuple(y1.tail(2).keys()), ('f6', 'f7'))

    #---------------------------------------------------------------------------
    def test_yarn_items_a(self) -> None:
        f1 = ff.parse('s(4,2)').rename('f1')
        f2 = ff.parse('s(4,5)').rename('f2')
        f3 = ff.parse('s(2,2)').rename('f3')
        f4 = ff.parse('s(2,8)').rename('f4')
        f5 = ff.parse('s(4,4)').rename('f5')
        f6 = ff.parse('s(6,4)').rename('f6')

        b1 = Bus.from_frames((f1, f2, f3))
        b2 = Bus.from_frames((f4, f5, f6))

        with temp_file('.zip') as fp1, temp_file('.zip') as fp2:
            b1.to_zip_pickle(fp1)
            b2.to_zip_pickle(fp2)

            bus_a = Bus.from_zip_pickle(fp1, max_persist=1).rename('a')
            bus_b = Bus.from_zip_pickle(fp2, max_persist=1).rename('b')

            y1 = Yarn.from_buses((bus_a, bus_b), retain_labels=False)

            labels = []
            for label, frame in y1.items():
                self.assertTrue(frame.__class__ is Frame)
                labels.append(label)

            self.assertEqual(labels, list(y1.index))
            self.assertEqual(y1.status['loaded'].sum(), 2)
            self.assertEqual(y1.status.loc[y1.status['loaded']].index.values.tolist(),
                ['f3', 'f6'])



    #---------------------------------------------------------------------------
    def test_yarn_values_a(self) -> None:
        f1 = ff.parse('s(4,2)').rename('f1')
        f2 = ff.parse('s(4,5)').rename('f2')
        f3 = ff.parse('s(2,2)').rename('f3')
        f4 = ff.parse('s(2,8)').rename('f4')
        f5 = ff.parse('s(4,4)').rename('f5')
        f6 = ff.parse('s(6,4)').rename('f6')

        b1 = Bus.from_frames((f1, f2, f3), name='a')
        b2 = Bus.from_frames((f4, f5, f6), name='b')

        y1 = Yarn.from_buses((b1, b2), retain_labels=False)
        s1 = y1.values
        self.assertEqual(len(s1), len(y1))
        self.assertEqual([f.shape for f in y1.values],
            [(4, 2), (4, 5), (2, 2), (2, 8), (4, 4), (6, 4)]
            )


    #---------------------------------------------------------------------------
    def test_yarn_display_a(self) -> None:
        f1 = ff.parse('s(4,2)').rename('f1')
        f2 = ff.parse('s(4,5)').rename('f2')
        f3 = ff.parse('s(2,2)').rename('f3')
        f4 = ff.parse('s(2,8)').rename('f4')
        f5 = ff.parse('s(4,4)').rename('f5')
        f6 = ff.parse('s(6,4)').rename('f6')

        b1 = Bus.from_frames((f1, f2, f3), name='a')
        b2 = Bus.from_frames((f4, f5, f6), name='b')

        y1 = Yarn.from_buses((b1, b2), retain_labels=False)

        d = y1.display(DisplayConfig(type_show=True, type_color=False))
        self.assertEqual(d.to_rows(),
            ['<Yarn>', '<Index>', 'f1      Frame', 'f2      Frame', 'f3      Frame', 'f4      Frame', 'f5      Frame', 'f6      Frame', '<<U2>   <object>'])

    #---------------------------------------------------------------------------
    def test_yarn_mloc_a(self) -> None:
        f1 = ff.parse('s(4,2)').rename('f1')
        f2 = ff.parse('s(4,5)').rename('f2')
        f3 = ff.parse('s(2,2)').rename('f3')
        f4 = ff.parse('s(2,8)').rename('f4')

        b1 = Bus.from_frames((f1, f2), name='a')
        b2 = Bus.from_frames((f3, f4), name='b')

        y1 = Yarn.from_buses((b1, b2), retain_labels=False)
        self.assertEqual(y1.mloc.shape, (4,))

    #---------------------------------------------------------------------------
    def test_yarn_dtypes_a(self) -> None:
        f1 = ff.parse('s(4,2)').rename('f1')
        f2 = ff.parse('s(4,5)').rename('f2')
        f3 = ff.parse('s(2,2)').rename('f3')
        f4 = ff.parse('s(2,8)').rename('f4')

        b1 = Bus.from_frames((f1, f2), name='a')
        b2 = Bus.from_frames((f3, f4), name='b')

        y1 = Yarn.from_buses((b1, b2), retain_labels=False)
        self.assertEqual(y1.dtypes.shape, (4, 8))

    #---------------------------------------------------------------------------
    def test_yarn_shapes_a(self) -> None:
        f1 = ff.parse('s(4,2)').rename('f1')
        f2 = ff.parse('s(4,5)').rename('f2')
        f3 = ff.parse('s(2,2)').rename('f3')
        f4 = ff.parse('s(2,8)').rename('f4')

        b1 = Bus.from_frames((f1, f2), name='a')
        b2 = Bus.from_frames((f3, f4), name='b')

        y1 = Yarn.from_buses((b1, b2), retain_labels=False)
        self.assertEqual(y1.shapes.to_pairs(),
                (('f1', (4, 2)), ('f2', (4, 5)), ('f3', (2, 2)), ('f4', (2, 8)))
                )

    #---------------------------------------------------------------------------
    def test_yarn_items_b(self) -> None:
        f1 = ff.parse('s(4,2)').rename('f1')
        f2 = ff.parse('s(4,5)').rename('f2')
        f3 = ff.parse('s(2,2)').rename('f3')
        f4 = ff.parse('s(2,8)').rename('f4')
        f5 = ff.parse('s(4,4)').rename('f5')
        f6 = ff.parse('s(6,4)').rename('f6')

        b1 = Bus.from_frames((f1, f2, f3))
        b2 = Bus.from_frames((f4, f5, f6))

        with temp_file('.zip') as fp1, temp_file('.zip') as fp2:
            b1.to_zip_pickle(fp1)
            b2.to_zip_pickle(fp2)

            bus_a = Bus.from_zip_pickle(fp1, max_persist=1).rename('a')
            bus_b = Bus.from_zip_pickle(fp2, max_persist=1).rename('b')

            y1 = Yarn.from_buses((bus_a, bus_b), retain_labels=False)

            s1 = y1.to_series()

            self.assertEqual(
                [(label, f.shape) for label, f in s1.items()],
                [('f1', (4, 2)), ('f2', (4, 5)), ('f3', (2, 2)), ('f4', (2, 8)), ('f5', (4, 4)), ('f6', (6, 4))]
                )


    #---------------------------------------------------------------------------
    def test_yarn_equals_a(self) -> None:
        f1 = ff.parse('s(4,2)').rename('f1')
        f2 = ff.parse('s(4,5)').rename('f2')
        f3 = ff.parse('s(2,2)').rename('f3')
        f4 = ff.parse('s(2,8)').rename('f4')

        b1 = Bus.from_frames((f1, f2), name='a')
        b2 = Bus.from_frames((f3, f4), name='b')

        y1 = Yarn.from_buses((b1, b2), retain_labels=False, name='foo')
        y2 = Yarn.from_buses((b1, b2), retain_labels=False, name='bar')

        self.assertTrue(y1.equals(y1))
        self.assertFalse(y1.equals(f1))
        self.assertFalse(y1.equals(f1, compare_class=True))

        self.assertTrue(y1.equals(y2))
        self.assertFalse(y1.equals(y2, compare_name=True))


    def test_yarn_equals_b(self) -> None:
        f1 = ff.parse('s(4,2)').rename('f1')
        f2 = ff.parse('s(4,5)').rename('f2')
        f3 = ff.parse('s(2,2)').rename('f3')
        f4 = ff.parse('s(2,8)').rename('f4')

        b1 = Bus.from_frames((f1, f2), name='a')
        b2 = Bus.from_frames((f3, f4), name='b')

        y1 = Yarn.from_buses((b1, b2), retain_labels=False)
        y2 = Yarn.from_buses((b1,), retain_labels=False)

        # fails on length
        self.assertFalse(y1.equals(y2))


    def test_yarn_equals_c(self) -> None:
        f1 = ff.parse('s(4,2)').rename('f1')
        f2 = ff.parse('s(4,5)').rename('f2')
        f3 = ff.parse('s(2,2)').rename('f3')
        f4a = ff.parse('s(2,8)').rename('f4a')
        f4b = ff.parse('s(2,8)').rename('f4b')

        b1 = Bus.from_frames((f1, f2), name='a')
        b2a = Bus.from_frames((f3, f4a), name='b')
        b2b = Bus.from_frames((f3, f4b), name='b')

        y1 = Yarn.from_buses((b1, b2a), retain_labels=False)
        y2 = Yarn.from_buses((b1, b2b), retain_labels=False)

        # fails on index
        self.assertFalse(y1.equals(y2))


    def test_yarn_equals_d(self) -> None:
        f1 = ff.parse('s(4,2)').rename('f1')
        f2 = ff.parse('s(4,5)').rename('f2')
        f3 = ff.parse('s(2,2)').rename('f3')
        f4a = ff.parse('s(2,8)').rename('f4')
        f4b = ff.parse('s(2,8)|v(str)').rename('f4')

        b1 = Bus.from_frames((f1, f2), name='a')
        b2a = Bus.from_frames((f3, f4a), name='b')
        b2b = Bus.from_frames((f3, f4b), name='b')

        y1 = Yarn.from_buses((b1, b2a), retain_labels=False)
        y2 = Yarn.from_buses((b1, b2b), retain_labels=False)

        # fails on index
        self.assertFalse(y1.equals(y2))

    def test_yarn_equals_e(self) -> None:
        f1 = ff.parse('s(4,2)').rename('f1')
        f2 = ff.parse('s(4,5)').rename('f2')
        f3 = ff.parse('s(2,2)').rename('f3')
        f4 = ff.parse('s(2,8)').rename('f4')
        f5 = ff.parse('s(4,4)').rename('f5')
        f6 = ff.parse('s(6,4)').rename('f6')

        b1 = Bus.from_frames((f1, f2, f3), name='a')
        b2 = Bus.from_frames((f4, f5, f6), name='b')

        y1 = Yarn.from_buses((b1, b2), retain_labels=True)

        with temp_file('.zip') as fp1, temp_file('.zip') as fp2:
            b1.to_zip_pickle(fp1)
            b2.to_zip_pickle(fp2)

            bus_a = Bus.from_zip_pickle(fp1, max_persist=1).rename('a')
            bus_b = Bus.from_zip_pickle(fp2, max_persist=1).rename('b')

            y2 = Yarn.from_buses((bus_a, bus_b), retain_labels=True)
            self.assertEqual(y2.status['loaded'].sum(), 0)

            self.assertTrue(y1.equals(y2))
            self.assertEqual(y2.status['loaded'].sum(), 2)


    #---------------------------------------------------------------------------
    def test_yarn_to_zip_pickle_a(self) -> None:
        f1 = ff.parse('s(4,2)').rename('f1')
        f2 = ff.parse('s(4,5)').rename('f2')
        f3 = ff.parse('s(2,2)').rename('f3')
        f4 = ff.parse('s(2,8)').rename('f4')
        f5 = ff.parse('s(4,4)').rename('f5')
        f6 = ff.parse('s(6,4)').rename('f6')

        b1 = Bus.from_frames((f1, f2, f3))
        b2 = Bus.from_frames((f4, f5, f6))

        with temp_file('.zip') as fp1, temp_file('.zip') as fp2, temp_file('.zip') as fp3:
            b1.to_zip_pickle(fp1)
            b2.to_zip_pickle(fp2)

            bus_a = Bus.from_zip_pickle(fp1, max_persist=1).rename('a')
            bus_b = Bus.from_zip_pickle(fp2, max_persist=1).rename('b')

            y1 = Yarn.from_buses((bus_a, bus_b), retain_labels=False)
            y1.to_zip_pickle(fp3)

            b3 = Bus.from_zip_pickle(fp3)
            self.assertTrue(b3.index.equals(y1.index))


    #---------------------------------------------------------------------------
    def test_yarn_iter_element_a(self) -> None:
        f1 = ff.parse('s(4,2)').rename('f1')
        f2 = ff.parse('s(4,5)').rename('f2')
        f3 = ff.parse('s(2,2)').rename('f3')
        f4 = ff.parse('s(2,8)').rename('f4')
        f5 = ff.parse('s(4,4)').rename('f5')
        f6 = ff.parse('s(6,4)').rename('f6')

        b1 = Bus.from_frames((f1, f2, f3), name='b1')
        b2 = Bus.from_frames((f4, f5, f6), name='b2')
        y1 = Yarn.from_buses((b1, b2), retain_labels=False)

        s1 = y1.iter_element().apply(lambda f: f.shape)
        self.assertEqual(s1.to_pairs(),
                (('f1', (4, 2)), ('f2', (4, 5)), ('f3', (2, 2)), ('f4', (2, 8)), ('f5', (4, 4)), ('f6', (6, 4))))

        self.assertEqual([f.name for f in y1.iter_element() if f.shape[0] > 2], #type: ignore
                ['f1', 'f2', 'f5', 'f6'],
                )

    #---------------------------------------------------------------------------
    def test_yarn_iter_element_items_a(self) -> None:
        f1 = ff.parse('s(4,2)').rename('f1')
        f2 = ff.parse('s(4,5)').rename('f2')
        f3 = ff.parse('s(2,2)').rename('f3')
        f4 = ff.parse('s(2,8)').rename('f4')
        f5 = ff.parse('s(4,4)').rename('f5')
        f6 = ff.parse('s(6,4)').rename('f6')

        b1 = Bus.from_frames((f1, f2, f3), name='b1')
        b2 = Bus.from_frames((f4, f5, f6), name='b2')
        y1 = Yarn.from_buses((b1, b2), retain_labels=False)

        s1 = y1.iter_element_items().apply(lambda label, f: (label, f.shape[0], f.shape[1]))

        self.assertEqual(s1.to_pairs(),
                (('f1', ('f1', 4, 2)), ('f2', ('f2', 4, 5)), ('f3', ('f3', 2, 2)), ('f4', ('f4', 2, 8)), ('f5', ('f5', 4, 4)), ('f6', ('f6', 6, 4)))
                )

        self.assertEqual([label for label, f in y1.iter_element_items() if f.shape[0] > 2],
                ['f1', 'f2', 'f5', 'f6'],
                )

    #---------------------------------------------------------------------------
    def test_yarn_drop_a(self) -> None:
        f1 = ff.parse('s(4,2)').rename('f1')
        f2 = ff.parse('s(4,5)').rename('f2')
        f3 = ff.parse('s(2,2)').rename('f3')
        f4 = ff.parse('s(2,8)').rename('f4')
        f5 = ff.parse('s(4,4)').rename('f5')
        f6 = ff.parse('s(6,4)').rename('f6')

        b1 = Bus.from_frames((f1, f2, f3), name='b1')
        b2 = Bus.from_frames((f4,), name='b2')
        b3 = Bus.from_frames((f5, f6), name='b3')

        y1 = Yarn.from_buses((b1, b2, b3), retain_labels=False)

        y2 = y1.drop['f3':'f5'] #type: ignore
        self.assertEqual(len(y2._series), 2) # 2 buses remain
        self.assertEqual([(f.name, f.shape) for f in y2.values],
                [('f1', (4, 2)), ('f2', (4, 5)), ('f6', (6, 4))]
                )

        y3 = y1.drop[y1.index.isin(('f1', 'f6'))]
        self.assertEqual([(f.name, f.shape) for f in y3.values],
                [('f2', (4, 5)), ('f3', (2, 2)), ('f4', (2, 8)), ('f5', (4, 4))]
                )

        y4 = y1.drop['f4']
        self.assertEqual([(f.name, f.shape) for f in y4.values],
                [('f1', (4, 2)), ('f2', (4, 5)), ('f3', (2, 2)), ('f5', (4, 4)), ('f6', (6, 4))]
                )

        y5 = y1.drop[['f4', 'f5', 'f6']]
        self.assertEqual([(f.name, f.shape) for f in y5.values],
                [('f1', (4, 2)), ('f2', (4, 5)), ('f3', (2, 2))]
                )

    def test_yarn_drop_b(self) -> None:
        f1 = ff.parse('s(4,2)').rename('f1')
        f2 = ff.parse('s(4,5)').rename('f2')
        f3 = ff.parse('s(2,2)').rename('f3')
        f4 = ff.parse('s(2,8)').rename('f4')
        f5 = ff.parse('s(4,4)').rename('f5')
        f6 = ff.parse('s(6,4)').rename('f6')

        b1 = Bus.from_frames((f1, f2, f3), name='b1')
        b2 = Bus.from_frames((f4,), name='b2')
        b3 = Bus.from_frames((f5, f6), name='b3')

        y1 = Yarn.from_buses((b1, b2, b3), retain_labels=False)

        y2 = y1.drop.iloc[2: 5]
        self.assertEqual(len(y2._series), 2) # 2 buses remain
        self.assertEqual([(f.name, f.shape) for f in y2.values],
                [('f1', (4, 2)), ('f2', (4, 5)), ('f6', (6, 4))]
                )

        y3 = y1.drop.iloc[np.array([True, False, False, False, False, True])]
        self.assertEqual([(f.name, f.shape) for f in y3.values],
                [('f2', (4, 5)), ('f3', (2, 2)), ('f4', (2, 8)), ('f5', (4, 4))]
                )

        y4 = y1.drop.iloc[3]
        self.assertEqual([(f.name, f.shape) for f in y4.values],
                [('f1', (4, 2)), ('f2', (4, 5)), ('f3', (2, 2)), ('f5', (4, 4)), ('f6', (6, 4))]
                )

        y5 = y1.drop.iloc[[3, 4, 5]]
        self.assertEqual([(f.name, f.shape) for f in y5.values],
                [('f1', (4, 2)), ('f2', (4, 5)), ('f3', (2, 2))]
                )

    #---------------------------------------------------------------------------
    def test_yarn_unpersist_a(self) -> None:
        f1 = ff.parse('s(4,2)').rename('f1')
        f2 = ff.parse('s(4,5)').rename('f2')
        f3 = ff.parse('s(2,2)').rename('f3')
        f4 = ff.parse('s(2,8)').rename('f4')
        f5 = ff.parse('s(4,4)').rename('f5')
        f6 = ff.parse('s(6,4)').rename('f6')

        b1 = Bus.from_frames((f1, f2, f3))
        b2 = Bus.from_frames((f4, f5, f6))

        with temp_file('.zip') as fp1, temp_file('.zip') as fp2:
            b1.to_zip_pickle(fp1)
            b2.to_zip_pickle(fp2)

            bus_a = Bus.from_zip_pickle(fp1, max_persist=1).rename('a')
            bus_b = Bus.from_zip_pickle(fp2, max_persist=1).rename('b')

            y1 = Yarn.from_buses((bus_a, bus_b), retain_labels=False)
            self.assertEqual(len(tuple(y1.items())), 6)

            self.assertEqual(y1.status['loaded'].sum(), 2)
            y1.unpersist()

            self.assertEqual(y1.status['loaded'].sum(), 0)
            self.assertEqual(len(tuple(y1.items())), 6)
            self.assertEqual(y1.status['loaded'].sum(), 2)



    #---------------------------------------------------------------------------
    def test_yarn_relabel_a(self) -> None:
        f1 = ff.parse('s(4,2)').rename('f1')
        f2 = ff.parse('s(4,5)').rename('f2')
        f3 = ff.parse('s(2,2)').rename('f3')
        f4 = ff.parse('s(2,8)').rename('f4')
        f5 = ff.parse('s(4,4)').rename('f5')
        f6 = ff.parse('s(6,4)').rename('f6')

        b1 = Bus.from_frames((f1, f2, f3))
        b2 = Bus.from_frames((f4,))
        b3 = Bus.from_frames((f5, f6))

        y1 = Yarn((b1, b2, b3))

        self.assertEqual(
                y1.relabel(lambda x: f'--{x}--').loc['--4--'].shape, (4, 4)
                )

        # None is a no-op
        self.assertEqual(
                y1.relabel(None).loc[4].shape, (4, 4)
                )

        with self.assertRaises(RelabelInvalid):
            y1.relabel({3,4,5})

        self.assertEqual(
                y1.relabel(tuple('abcdef'))['d':].status['shape'].to_pairs(),
                (('d', (2, 8)), ('e', (4, 4)), ('f', (6, 4)))
                )


        y2 = Yarn((b1, b2, b3), index=tuple('abcdef'))
        self.assertEqual(y2.index.values.tolist(), ['a', 'b', 'c', 'd', 'e', 'f'])
        self.assertEqual(y2.relabel(IndexAutoFactory).index.values.tolist(), [0, 1, 2, 3, 4, 5])


    #---------------------------------------------------------------------------
    def test_yarn_relabel_flat_a(self) -> None:
        f1 = ff.parse('s(4,2)').rename('f1')
        f2 = ff.parse('s(4,5)').rename('f2')
        f3 = ff.parse('s(2,2)').rename('f3')
        f4 = ff.parse('s(2,8)').rename('f4')
        f5 = ff.parse('s(4,4)').rename('f5')
        f6 = ff.parse('s(6,4)').rename('f6')

        b1 = Bus.from_frames((f1, f2, f3))
        b2 = Bus.from_frames((f4,))
        b3 = Bus.from_frames((f5, f6))

        y1 = Yarn((b1, b2, b3), index=IndexHierarchy.from_product(('a', 'b'), (1, 2, 3)))

        self.assertEqual(
                y1.relabel_flat()[('a', 3):].status['shape'].to_pairs(),
                ((('a', 3), (2, 2)), (('b', 1), (2, 8)), (('b', 2), (4, 4)), (('b', 3), (6, 4)))
                )

    def test_yarn_relabel_flat_b(self) -> None:
        f1 = ff.parse('s(4,2)').rename('f1')
        f2 = ff.parse('s(4,5)').rename('f2')
        f3 = ff.parse('s(2,2)').rename('f3')

        b1 = Bus.from_frames((f1, f2, f3))

        y1 = Yarn((b1,))
        with self.assertRaises(RuntimeError):
            y1.relabel_flat()

    #---------------------------------------------------------------------------
    def test_yarn_relabel_level_add_a(self) -> None:
        f1 = ff.parse('s(4,2)').rename('f1')
        f2 = ff.parse('s(4,5)').rename('f2')
        f3 = ff.parse('s(2,2)').rename('f3')
        f4 = ff.parse('s(2,8)').rename('f4')
        f5 = ff.parse('s(4,4)').rename('f5')
        f6 = ff.parse('s(6,4)').rename('f6')

        b1 = Bus.from_frames((f1, f2, f3))
        b2 = Bus.from_frames((f4,))
        b3 = Bus.from_frames((f5, f6))

        y1 = Yarn((b1, b2, b3), index=IndexHierarchy.from_product(('a', 'b'), (1, 2, 3)))

        self.assertEqual(
                y1.relabel_level_add('c').iloc[4:].status['shape'].to_pairs(),
                ((('c', 'b', 2), (4, 4)), (('c', 'b', 3), (6, 4)))
                )

    #---------------------------------------------------------------------------
    def test_yarn_relabel_level_drop_a(self) -> None:
        f1 = ff.parse('s(4,2)').rename('f1')
        f2 = ff.parse('s(4,5)').rename('f2')
        f3 = ff.parse('s(2,2)').rename('f3')
        f4 = ff.parse('s(2,8)').rename('f4')
        f5 = ff.parse('s(4,4)').rename('f5')
        f6 = ff.parse('s(6,4)').rename('f6')

        b1 = Bus.from_frames((f1, f2, f3))
        b2 = Bus.from_frames((f4,))
        b3 = Bus.from_frames((f5, f6))

        y1 = Yarn((b1, b2, b3), index=IndexHierarchy.from_product(('a', 'b'), (1, 2, 3)))

        self.assertEqual(
                y1.iloc[[0,2,4]].relabel_level_drop(1).status['shape'].to_pairs(),
                ((1, (4, 2)), (3, (2, 2)), (2, (4, 4)))
                )

    def test_yarn_relabel_level_drop_b(self) -> None:
        f1 = ff.parse('s(4,2)').rename('f1')
        f2 = ff.parse('s(4,5)').rename('f2')
        f3 = ff.parse('s(2,2)').rename('f3')

        b1 = Bus.from_frames((f1, f2, f3))

        y1 = Yarn((b1,))
        with self.assertRaises(RuntimeError):
            y1.relabel_level_drop()



    #---------------------------------------------------------------------------
    def test_yarn_rehierarch_a(self) -> None:
        f1 = ff.parse('s(4,2)').rename('f1')
        f2 = ff.parse('s(4,5)').rename('f2')
        f3 = ff.parse('s(2,2)').rename('f3')
        f4 = ff.parse('s(2,8)').rename('f4')
        f5 = ff.parse('s(4,4)').rename('f5')
        f6 = ff.parse('s(6,4)').rename('f6')

        b1 = Bus.from_frames((f1, f2, f3))
        b2 = Bus.from_frames((f4,))
        b3 = Bus.from_frames((f5, f6))

        y1 = Yarn((b1, b2, b3), index=IndexHierarchy.from_product(('a', 'b'), (1, 2, 3)))
        self.assertEqual(
                y1.iloc[[0,2,4]].rehierarch((1, 0)).status['shape'].to_pairs(),
                (((1, 'a'), (4, 2)), ((3, 'a'), (2, 2)), ((2, 'b'), (4, 4)))
                )

    def test_yarn_rehierarch_b(self) -> None:
        f1 = ff.parse('s(4,2)').rename('f1')
        f2 = ff.parse('s(4,5)').rename('f2')
        f3 = ff.parse('s(2,2)').rename('f3')

        b1 = Bus.from_frames((f1, f2, f3))

        y1 = Yarn((b1,))
        with self.assertRaises(RuntimeError):
            y1.rehierarch((1,0))




if __name__ == '__main__':
    unittest.main()







































import typing as tp
import unittest

import numpy as np

# from hypothesis import strategies as st
from hypothesis import given
# from hypothesis import reproduce_failure  # type: ignore

from static_frame.test.property.strategies import get_labels
from static_frame.test.property.strategies import get_index_any

from static_frame.test.test_case import TestCase

from static_frame import Index
from static_frame import IndexGO
# from static_frame import Series
# from static_frame import Frame
# from static_frame import FrameGO
# from static_frame import TypeBlocks
# from static_frame import Display



class TestUnit(TestCase):


    @given(get_labels())
    def test_index_values_len(self, values: tp.Sequence[tp.Hashable]) -> None:

        def property_values(cls: tp.Type[Index], values: tp.Sequence[tp.Hashable]) -> None:
            #Property: that the length of the index is the length of the (unique) values.
            index = cls(values)
            self.assertEqual(len(index), len(values))
            self.assertEqual(len(index.values), len(values))

        property_values(Index, values)
        property_values(IndexGO, values)

    @given(get_labels())
    def test_index_values_list(self, values: tp.Sequence[tp.Hashable]) -> None:

        def property_values(cls: tp.Type[Index], values: tp.Iterable[tp.Hashable]) -> None:
            index = cls(values)
            # must cast both sides to the dtype, as some int to float conversions result in different floats
            self.assertAlmostEqualValues(index.values, np.array(values, dtype=index.values.dtype))

        property_values(Index, values)
        property_values(IndexGO, values)


    @given(get_labels())
    def test_index_loc_to_iloc_element(self, values: tp.Sequence[tp.Hashable]) -> None:

        def property_loc_to_iloc_element(cls: tp.Type[Index], values: tp.Iterable[tp.Hashable]) -> None:
            index = cls(values)
            for i, v in enumerate(values):
                self.assertEqual(index._loc_to_iloc(v), i)

        property_loc_to_iloc_element(Index, values)
        property_loc_to_iloc_element(IndexGO, values)

    @given(get_labels(min_size=1))
    def test_index_loc_to_iloc_slice(self, values: tp.Sequence[tp.Hashable]) -> None:

        def property_loc_to_iloc_slice(cls: tp.Type[Index], values: tp.Iterable[tp.Hashable]) -> None:
            # Property: that the key translates to the appropriate slice.
            index = cls(values)
            for i, v in enumerate(values):
                # insure that we get teh same slice going through loc that we would get by direct iloc
                if v is None:
                    self.assertEqual(index._loc_to_iloc(slice(v, None)), slice(None)) #type: ignore [unreachable]
                else:
                    self.assertEqual(index._loc_to_iloc(slice(v, None)), slice(i, None))

        property_loc_to_iloc_slice(Index, values)
        property_loc_to_iloc_slice(IndexGO, values)



    @given(get_labels(min_size=2))
    def test_index_go_append(self, values: tp.Sequence[tp.Hashable]) -> None:

        index = IndexGO(values[:-1])
        length_start = len(index)
        index.append(values[-1])
        length_end = len(index)
        self.assertEqual(length_start + 1, length_end)


    @given(get_labels(min_size=1))
    def test_index_isin(self, labels: tp.Sequence[tp.Hashable]) -> None:
        index = Index(labels)
        self.assertTrue(index.isin((labels[0],))[0])


    #---------------------------------------------------------------------------
    @given(get_index_any())
    def test_index_display(self, index: Index) -> None:

        d1 = index.display()
        self.assertTrue(len(d1) > 0)

        d2 = index.display_tall()
        self.assertTrue(len(d2) > 0)

        d3 = index.display_wide()
        self.assertTrue(len(d3) > 0)

    @given(get_index_any())
    def test_index_to_series(self, index: Index) -> None:
        s1 = index.to_series()
        self.assertEqual(len(s1), len(index))

if __name__ == '__main__':
    unittest.main()



import typing as tp
import unittest

import numpy as np
import frame_fixtures as ff

from static_frame.test.test_case import TestCase
from static_frame.core.bus import Bus
from static_frame.core.frame import Frame
from static_frame.core.index import Index
from static_frame.core.index_level import TreeNodeT
from static_frame.core.index_hierarchy import IndexHierarchy
from static_frame.core.exception import ErrorInitBus
from static_frame.core.exception import ErrorInitQuilt
from static_frame.core.exception import AxisInvalid
from static_frame.core.exception import ErrorInitYarn

from static_frame.core.axis_map import bus_to_hierarchy
from static_frame.core.axis_map import buses_to_hierarchy


class TestUnit(TestCase):

    def test_axis_map_b(self) -> None:

        f1 = ff.parse('s(4,4)|v(int,float)').rename('f1')
        f2 = ff.parse('s(4,4)|v(str)').rename('f2')
        f3 = ff.parse('s(4,4)|v(bool)').rename('f3')

        b1 = Bus.from_frames((f1, f2, f3))

        with self.assertRaises(AxisInvalid):
            bus_to_hierarchy(b1, deepcopy_from_bus=False, axis=3, init_exception_cls=ErrorInitQuilt)

    def test_index_map_a(self) -> None:

        f1 = ff.parse('s(4,4)|v(int,float)').rename('f1')
        f2 = ff.parse('s(4,4)|v(str)').rename('f2')
        f3 = ff.parse('s(4,4)|v(bool)').rename('f3')
        b1 = Bus.from_frames((f1, f2, f3), name='a')

        f4 = ff.parse('s(4,4)|v(int,float)').rename('f4')
        f5 = ff.parse('s(4,4)|v(str)').rename('f5')
        b2 = Bus.from_frames((f4, f5), name='b')

        post = buses_to_hierarchy((b1, b2), (b1.name, b2.name), deepcopy_from_bus=False, init_exception_cls=ErrorInitYarn)

        self.assertEqual(post.values.tolist(),
                [['a', 'f1'], ['a', 'f2'], ['a', 'f3'], ['b', 'f4'], ['b', 'f5']])

    def compare_trees(self, tree1: TreeNodeT, tree2: TreeNodeT) -> None:
        self.assertSequenceEqual(tree1.keys(), tree2.keys())

        for k, v1 in tree1.items():
            v2 = tree2[k]
            if isinstance(v1, Index):
                self.assertTrue(v1.equals(v2))
            else:
                self.compare_trees(v1, v2)

    def test_bus_to_hierarchy_a(self) -> None:
        f1 = ff.parse('s(4,4)|v(int,float)|c(I, str)').rename('f1')
        f2 = ff.parse('s(4,4)|v(str)|c(I, str)').rename('f2')
        f3 = ff.parse('s(4,4)|v(bool)|c(I, str)').rename('f3')
        b1 = Bus.from_frames((f1, f2, f3), name='a')

        indices = Index((0, 1, 2, 3))
        columns = Index(("zZbu", "ztsv", "zUvW", "zkuW"))

        for _, frame in b1.items():
            self.assertTrue(indices.equals(frame.index))
            self.assertTrue(columns.equals(frame.columns))


        def test_assertions(axis: int, flag: bool) -> None:
            hierarchy, opposite = bus_to_hierarchy(b1, axis=axis, deepcopy_from_bus=flag, init_exception_cls=ErrorInitBus)

            if axis == 0:
                expected_tree: tp.Dict[str, Index] = {
                    'f1': indices, 'f2': indices, 'f3': indices
                }
                expected_index = columns
            else:
                expected_index = indices
                expected_tree = {'f1': columns, 'f2': columns, 'f3': columns}

            self.compare_trees(hierarchy.to_tree(), expected_tree)
            self.assertTrue(expected_index.equals(opposite))

        for axis in (0, 1):
            for flag in (True, False):
                test_assertions(axis, flag)

    def test_bus_to_hierarchy_b(self) -> None:

        class CustomError(Exception):
            pass

        tree1 = dict(a_I=Index((1,2,3)), a_II=Index((1,2,3)))
        tree2 = dict(b_I=Index((1,2,3)), b_II=Index((1,2,3)))
        tree3 = dict(c_I=Index((1,2,3)), c_II=Index((1,2,3)))
        index1 = IndexHierarchy.from_tree(tree1)
        index2 = IndexHierarchy.from_tree(tree2)
        index3 = IndexHierarchy.from_tree(tree3)
        values = np.arange(36).reshape(6,6)

        # Align all the frames on columns!
        f1 = Frame(values, index=index1, columns=index1, name="f1")
        f2 = Frame(values, index=index2, columns=index1, name="f2")
        f3 = Frame(values, index=index3, columns=index1, name="f3")
        b1 = Bus.from_frames((f1, f2, f3))

        def test_assertions(hierarchy: IndexHierarchy, opposite: Index) -> None:
            expected_tree = dict(f1=tree1, f2=tree2, f3=tree3)
            self.compare_trees(hierarchy.to_tree(), expected_tree)
            self.assertTrue(index1.equals(opposite))

        test_assertions(*bus_to_hierarchy(b1, axis=0, deepcopy_from_bus=False, init_exception_cls=CustomError))
        test_assertions(*bus_to_hierarchy(b1, axis=0, deepcopy_from_bus=True, init_exception_cls=CustomError))

        # Cannot do this since the frames do not share the same index
        with self.assertRaises(CustomError):
            bus_to_hierarchy(b1, axis=1, deepcopy_from_bus=False, init_exception_cls=CustomError)

        with self.assertRaises(CustomError):
            bus_to_hierarchy(b1, axis=1, deepcopy_from_bus=True, init_exception_cls=CustomError)

        # Align all the frames on index!
        f1 = Frame(values, index=index1, columns=index1, name="f1")
        f2 = Frame(values, index=index1, columns=index2, name="f2")
        f3 = Frame(values, index=index1, columns=index3, name="f3")
        b1 = Bus.from_frames((f1, f2, f3))

        test_assertions(*bus_to_hierarchy(b1, axis=1, deepcopy_from_bus=False, init_exception_cls=CustomError))
        test_assertions(*bus_to_hierarchy(b1, axis=1, deepcopy_from_bus=True, init_exception_cls=CustomError))

        # Cannot do this since the frames do not share the same columns
        with self.assertRaises(CustomError):
            bus_to_hierarchy(b1, axis=0, deepcopy_from_bus=False, init_exception_cls=CustomError)

        with self.assertRaises(CustomError):
            bus_to_hierarchy(b1, axis=0, deepcopy_from_bus=True, init_exception_cls=CustomError)

    def test_buses_to_hierarchy_a(self) -> None:
        f1 = ff.parse('s(4,4)|v(int,float)').rename('f1')
        f2 = ff.parse('s(4,4)|v(str)').rename('f2')
        f3 = ff.parse('s(4,4)|v(bool)').rename('f3')
        b1 = Bus.from_frames((f1, f2, f3), name='a')

        f4 = ff.parse('s(4,4)|v(int,float)').rename('f4')
        f5 = ff.parse('s(4,4)|v(str)').rename('f5')
        b2 = Bus.from_frames((f4, f5), name='b')

        post = buses_to_hierarchy((b1, b2), (b1.name, b2.name), deepcopy_from_bus=False, init_exception_cls=ErrorInitYarn)

        self.assertEqual(post.values.tolist(),
                [['a', 'f1'], ['a', 'f2'], ['a', 'f3'], ['b', 'f4'], ['b', 'f5']])

    def test_buses_to_hierarchy_b(self) -> None:
        f1 = ff.parse('s(4,4)|v(int,float)').rename('f1')
        f2 = ff.parse('s(4,4)|v(str)').rename('f2')
        f3 = ff.parse('s(4,4)|v(bool)').rename('f3')
        b1 = Bus.from_frames((f1, f2, f3), name='foo')

        f4 = ff.parse('s(4,4)|v(int,float)').rename('f4')
        f5 = ff.parse('s(4,4)|v(str)').rename('f5')
        b2 = Bus.from_frames((f4, f5), name='foo')

        with self.assertRaises(ErrorInitYarn):
            _ = buses_to_hierarchy((b1, b2),
                    (b1.name, b2.name),
                    deepcopy_from_bus=False,
                    init_exception_cls=ErrorInitYarn)


if __name__ == '__main__':
    unittest.main()



# import typing as tp
import unittest
import operator

# from hypothesis import strategies as st
from hypothesis import given
from arraykit import isna_element

from static_frame.core.interface import UFUNC_UNARY_OPERATORS
from static_frame.core.interface import UFUNC_BINARY_OPERATORS
from static_frame.core.interface import UFUNC_AXIS_SKIPNA

from static_frame.test.property import strategies as sfst
from static_frame.test.test_case import TestCase

from static_frame import Series



class TestUnit(TestCase):


    @given(sfst.get_series())
    def test_basic_attributes(self, s1: Series) -> None:

        self.assertEqual(s1.dtype, s1.values.dtype)
        # self.assertEqual(s1.shape, s1.shape)
        self.assertEqual(s1.ndim, 1)

        if s1.shape[0] > 0:
            self.assertTrue(s1.size > 0)
            self.assertTrue(s1.nbytes > 0)


    @given(sfst.get_series(dtype_group=sfst.DTGroup.NUMERIC, min_size=1))
    def test_unary_operators_numeric(self, s1: Series) -> None:
        for op in UFUNC_UNARY_OPERATORS:
            if op == '__invert__': # invalid on non Boolean
                continue
            func = getattr(operator, op)
            a = func(s1).values
            b = func(s1.values)
            self.assertAlmostEqualArray(a, b)


    @given(sfst.get_series(dtype_group=sfst.DTGroup.BOOL, min_size=1))
    def test_unary_operators_boolean(self, s1: Series) -> None:
        for op in UFUNC_UNARY_OPERATORS:
            if op != '__invert__': # valid on Boolean
                continue
            func = getattr(operator, op)
            a = func(s1).values
            b = func(s1.values)
            self.assertAlmostEqualArray(a, b)



    @given(sfst.get_series(dtype_group=sfst.DTGroup.NUMERIC))
    def test_binary_operators_numeric(self, s1: Series) -> None:
        for op in UFUNC_BINARY_OPERATORS:
            if op in {
                    '__matmul__',
                    '__pow__',
                    '__lshift__',
                    '__rshift__',
                    '__and__',
                    '__xor__',
                    '__or__',
                    '__mod__',
                    }:
                continue
            func = getattr(operator, op)
            values = s1.values
            a = func(s1, s1).values
            b = func(values, values)
            self.assertAlmostEqualArray(a, b)

    @given(sfst.get_series(dtype_group=sfst.DTGroup.BOOL))
    def test_binary_operators_boolean(self, s1: Series) -> None:
        for op in UFUNC_BINARY_OPERATORS:
            if op not in {
                    '__and__',
                    '__xor__',
                    '__or__',
                    }:
                continue
            func = getattr(operator, op)
            values = s1.values
            a = func(s1, s1).values
            b = func(values, values)
            self.assertAlmostEqualArray(a, b)


    @given(sfst.get_series(dtype_group=sfst.DTGroup.NUMERIC, min_size=1))
    def test_ufunc_axis(self, s1: Series) -> None:
        for attr, attrs in UFUNC_AXIS_SKIPNA.items():
            a = getattr(s1, attr)()
            func = attrs.ufunc_skipna
            b = func(s1.values)
            self.assertEqualWithNaN(a, b)

    @given(sfst.get_series(min_size=1))
    def test_isin(self, s1: Series) -> None:

        value = s1.iloc[0]
        if not isna_element(value):
            self.assertTrue(s1.isin((value,)).iloc[0])


#     @given(sfst.get_series(min_size=1), sfst.get_series(min_size=1), sfst.get_series(min_size=1))
#     def test_from_overlay(self,
#                 s1: Series,
#                 s2: Series,
#                 s3: Series,
#                 ) -> None:

#         # NOTE: this fails dues to numpy doing this:
# #         In : np.array(9007199268722005).astype(float).tolist()
#  # 9007199268722004.0
#         # this happens in calls to np.union1d

#         post = Series.from_overlay((s1, s2, s3))
#         self.assertTrue(post.index.equals(s1.index.union(s2.index, s3.index)))





if __name__ == '__main__':
    unittest.main()



import os
# import typing as tp

from doc.source.conf import get_jinja_contexts


HEADER = '.. NOTE: auto-generated file, do not edit'


def get_rst_import(group: str, name: str) -> str:
    '''
    This approach works when we can import macros in this location. This does not yet work with ReadTheDocs.
    '''
    return f'''
.. jinja:: ctx

    {{% import 'doc/source/macros.jinja' as macros %}}

    {{{{ macros.{group}(examples_defined=examples_defined, *interface['{name}']) }}}}

'''

def get_rst_embed(group: str, name: str) -> str:
    '''
    This approach does not require importing the macro, and works with ReadTheDocs.
    '''
    doc_dir = os.path.abspath(os.path.dirname(__file__))
    fp = os.path.join(doc_dir, 'source', 'macros.jinja')

    with open(fp) as f:
        macro_lines = f.readlines()

    lines = [HEADER]
    lines.append('''
.. jinja:: ctx
''')

    for line in macro_lines:
        lines.append('    ' + line.rstrip())

    lines.append(f'''
    {{{{ {group}(examples_defined=examples_defined, *interface['{name}']) }}}}
''')
    return '\n'.join(lines)


def source_build() -> None:

    doc_dir = os.path.abspath(os.path.dirname(__file__))

    contexts = get_jinja_contexts()

    for group in ('api_detail', 'api_overview'):
        group_dir = os.path.join(doc_dir, 'source', group)

        for name, cls, frame in contexts['interface'].values():

            name_chars = []
            last_is_upper = False
            for i, char in enumerate(name):
                if char.isupper():
                    if i != 0 and not last_is_upper:
                        name_chars.append('_')
                    name_chars.append(char.lower())
                    last_is_upper = True
                else:
                    name_chars.append(char)
                    last_is_upper = False

            file_name = ''.join(name_chars) + '.rst'
            fp = os.path.join(group_dir, file_name)

            if not os.path.exists(fp):
                raise RuntimeError(f'must create and add RST file {fp}')

            print(fp)
            rst = get_rst_embed(group, name)
            # rst = get_rst_import(group, name)

            # print(rst)
            with open(fp, 'w') as f:
                f.write(rst)



if __name__ == '__main__':

    source_build()


import unittest

import frame_fixtures as ff

from static_frame.test.test_case import TestCase
from static_frame.core.index import Index
from static_frame.core.index_hierarchy import IndexHierarchy
from static_frame.core.pivot import pivot_records_items
# from static_frame.core.pivot import pivot_core

class TestUnit(TestCase):

    def test_pivot_records_items_a(self) -> None:
        frame = ff.parse('s(3,6)|v(int,str,bool)|c(I,str)|i(I,int)')
        group_fields = ['zUvW',] # needs to be valif loc selection
        group_depth = 1
        data_fields = ['zkuW', 'z2Oo']
        func_single = sum
        func_map = None
        loc_to_iloc = frame.columns.loc_to_iloc
        post = tuple(pivot_records_items(
                blocks=frame._blocks,
                group_fields_iloc=loc_to_iloc(group_fields),
                group_depth=group_depth,
                data_fields_iloc=loc_to_iloc(data_fields),
                func_single=func_single,
                func_map=func_map,
                ))
        self.assertEqual(post,
                ((False, [201945, 1]), (True, [129017, False])))


    def test_pivot_records_items_b(self) -> None:
        frame = ff.parse('s(3,6)|v(int,str,bool)|c(I,str)|i(I,int)')
        group_fields = ['zUvW',] # needs to be valif loc selection
        group_depth = 1
        data_fields = ['zkuW', 'z2Oo']
        func_single = None
        func_map = (('zkuW', sum), ('z2Oo', min))
        loc_to_iloc = frame.columns.loc_to_iloc

        post = tuple(pivot_records_items(
                blocks=frame._blocks,
                group_fields_iloc=loc_to_iloc(group_fields),
                group_depth=group_depth,
                data_fields_iloc=loc_to_iloc(data_fields),
                func_single=func_single,
                func_map=func_map,
                ))
        self.assertEqual(post,
                ((False, [201945, 35021, 1, False]),
                (True, [129017, 129017, False, False]))
                )


    def test_pivot_core_a(self) -> None:

        frame = ff.parse('s(20,4)|v(int)').assign[0].apply(lambda s: s % 4).assign[1].apply(lambda s: s % 3)

        # by default we get a tuple index
        post1 = frame.pivot([0, 1])
        self.assertEqual(post1.index.name, (0, 1))
        self.assertIs(post1.index.__class__, Index)
        self.assertTrue(post1.to_pairs(),
                ((2, (((0, 0), 463099), ((0, 1), -88017), ((0, 2), 35021), ((1, 0), 92867), ((1, 2), 96520), ((2, 0), 172133), ((2, 1), 279191), ((2, 2), 13448), ((3, 0), 255338), ((3, 1), 372807), ((3, 2), 155574))), (3, (((0, 0), 348362), ((0, 1), 175579), ((0, 2), 105269), ((1, 0), 58768), ((1, 2), 13448), ((2, 0), 84967), ((2, 1), 239151), ((2, 2), 170440), ((3, 0), 269300), ((3, 1), 204528), ((3, 2), 493169))))
                )

        # can provide index constructor
        post2 = frame.pivot([0, 1],
                index_constructor=IndexHierarchy.from_labels)
        self.assertEqual(post2.index.name, (0, 1))
        self.assertIs(post2.index.__class__, IndexHierarchy)
        self.assertTrue(post2.to_pairs(),
                ((2, (((0, 0), 463099), ((0, 1), -88017), ((0, 2), 35021), ((1, 0), 92867), ((1, 2), 96520), ((2, 0), 172133), ((2, 1), 279191), ((2, 2), 13448), ((3, 0), 255338), ((3, 1), 372807), ((3, 2), 155574))), (3, (((0, 0), 348362), ((0, 1), 175579), ((0, 2), 105269), ((1, 0), 58768), ((1, 2), 13448), ((2, 0), 84967), ((2, 1), 239151), ((2, 2), 170440), ((3, 0), 269300), ((3, 1), 204528), ((3, 2), 493169))))
                )


if __name__ == '__main__':
    unittest.main()






import timeit
from datetime import date
from datetime import datetime
from datetime import timedelta
import calendar

import typing as tp
from static_frame.core.display_color import HexColor

import numpy as np
dt64 = np.datetime64

import static_frame as sf


def get_date_strings():
    # get 30 years of daily data, 10957 observations
    post = sf.IndexDate.from_date_range('1990-01-01', '2019-12-31')
    return post.values.astype(str).tolist()


class DirectFromString:
    NUMBER = 500
    def __init__(self):
        self.date_str = get_date_strings()

    def dt(self):
        date_str = self.date_str
        array = np.empty(len(date_str), dtype=object)
        for i, msg in enumerate(date_str):
            array[i] = date.fromisoformat(msg)
        return array

    def dt64(self):
        return np.array(self.date_str, dtype='datetime64[D]')
        # return np.array(self.date_str, dtype=np.datetime64)


class ParseFromString:
    NUMBER = 50

    def __init__(self):
        self.date_str = get_date_strings()

    def dt(self):
        date_str = self.date_str
        array = np.empty(len(date_str), dtype=object)
        for i, msg in enumerate(date_str):
            array[i] = datetime.strptime(msg, '%Y-%m-%d').date()
        return array

    def dt64(self):
        return np.array(self.date_str, dtype=np.datetime64)

class ShiftDay:
    NUMBER = 500

    def __init__(self):
        runner = DirectFromString()
        self.array_dt64 = runner.dt64()
        self.array_dt = runner.dt()

    def dt(self):
        td = timedelta(days=1)
        np.array([d + td for d in self.array_dt], dtype=object)

    def dt64(self):
        self.array_dt64 + 1

class DeriveYear:
    NUMBER = 100

    def __init__(self):
        runner = DirectFromString()
        self.array_dt64 = runner.dt64()
        self.array_dt = runner.dt()

    def dt(self):
        np.array([x.year for x in self.array_dt], dtype=int)

    def dt64(self):
        self.array_dt64.astype('datetime64[Y]')

class TrueOnMonday:
    NUMBER = 100

    def __init__(self):
        runner = DirectFromString()
        self.array_dt64 = runner.dt64()
        self.array_dt = runner.dt()

    def dt(self):
        np.array([x.weekday() == 0 for x in self.array_dt], dtype=bool)

    def dt64(self):
        np.array(
            ((self.array_dt64.astype(int) + 3) % 7) == 0,
            dtype=int,
            )

class TrueOnEOM:
    NUMBER = 100

    def __init__(self):
        runner = DirectFromString()
        self.array_dt64 = runner.dt64()
        self.array_dt = runner.dt()

    def dt(self):

        np.array([d.day == calendar.monthrange(d.year, d.month)[1] for d in self.array_dt], dtype=bool)

    def dt64(self):
        np.array(
            (self.array_dt64.astype('datetime64[M]') + 1).astype('datetime64[D]') - 1 == self.array_dt64,
            dtype=bool,
            )


class TrueLessThan:
    NUMBER = 400

    def __init__(self):
        runner = DirectFromString()
        self.array_dt64 = runner.dt64()
        self.boundary_dt64 = np.datetime64('2005-06-18')
        self.array_dt = runner.dt()
        self.boundary_dt = date(2005, 6, 18)


    def dt(self):
        self.array_dt < self.boundary_dt

    def dt64(self):
        self.array_dt64 < self.boundary_dt64


class ConvertToStr:
    NUMBER = 100

    def __init__(self):
        runner = DirectFromString()
        self.array_dt64 = runner.dt64()
        self.array_dt = runner.dt()

    def dt(self):
        np.array([str(d) for d in self.array_dt], dtype=object)

    def dt64(self):
        self.array_dt64.astype(str)



def get_format():

    name_root_last = None
    name_root_count = 0

    def format(key: tp.Tuple[tp.Any, str], v: object) -> str:
        nonlocal name_root_last
        nonlocal name_root_count

        if isinstance(v, float):
            if np.isnan(v):
                return ''
            return str(round(v, 4))
        if isinstance(v, (bool, np.bool_)):
            if v:
                return HexColor.format_terminal('green', str(v))
            return HexColor.format_terminal('orange', str(v))

        return str(v)

    return format



def run_test():
    records = []
    for cls in (
            DirectFromString,
            ParseFromString,
            ShiftDay,
            DeriveYear,
            TrueOnMonday,
            TrueOnEOM,
            TrueLessThan,
            ConvertToStr,
            ):
        runner = cls()
        record = [cls.__name__, cls.NUMBER]
        for func_name in ('dt', 'dt64'):
            result = timeit.timeit(
                    f'runner.{func_name}()',
                    globals=locals(),
                    number=cls.NUMBER)
            record.append(result)
        records.append(record)

    f = sf.FrameGO.from_records(records,
            columns=('name', 'number', 'dt', 'dt64')
            )

    f['dt64/dt'] = f['dt64'] / f['dt']
    f['dt/dt64'] = f['dt'] / f['dt64']

    f['dt64-faster'] = f['dt/dt64'] > 1

    display = f.iter_element_items().apply(get_format())

    config = sf.DisplayConfig(
            cell_max_width_leftmost=np.inf,
            cell_max_width=np.inf,
            type_show=False,
            display_rows=200,
            include_index=False,
            )
    print(display.display(config))
    # import ipdb; ipdb.set_trace()

if __name__ == '__main__':
    run_test()


    # print(len(get_date_strings()))


    # ptest = ArrayCreationFromString()
    # post = ptest.dt()
    # post = ptest.dt64()

    # import ipdb; ipdb.set_trace()






import unittest

import numpy as np

from static_frame.test.test_case import TestCase
from static_frame.core.frame import Frame
from static_frame.test.test_case import temp_file

class TestUnit(TestCase):

    def test_frame_to_npz_a(self) -> None:
        # this forced a RuntimeError: File size unexpectedly exceeded ZIP64 limit error until the `force_zip64` parameter was added
        a1 = np.arange(15_000 * 18_000).reshape(15_000, 18_000)
        f1 = Frame(a1)

        with temp_file('.npz') as fp:
            f1.to_npz(fp)
            # f2 = Frame.from_npz(fp)
            # f1.equals(f2, compare_dtype=True, compare_class=True, compare_name=True)



if __name__ == '__main__':
    unittest.main()

